(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var _ = Package.underscore._;
var Random = Package.random.Random;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var SandstormBackend = Package['sandstorm-backend'].SandstormBackend;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var check = Package.check.check;
var Match = Package.check.Match;

(function(){

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/accounts-identity/accounts-identity-server.js                                             //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
function linkIdentityToAccountInternal(db, backend, identityId, accountId) {                          // 17
  // Links the identity to the account. If the account is a demo account, makes the account durable   // 18
  // and gives the identity login access to it.                                                       // 19
                                                                                                      // 20
  check(db, SandstormDb);                                                                             // 21
  check(backend, SandstormBackend);                                                                   // 22
  check(identityId, String);                                                                          // 23
  check(accountId, String);                                                                           // 24
                                                                                                      // 25
  var accountUser = Meteor.users.findOne({_id: accountId});                                           // 26
  if (!accountUser) {                                                                                 // 27
    throw new Meteor.Error(404, "No account found with ID " + accountId);                             // 28
  }                                                                                                   // 29
                                                                                                      // 30
  if (accountUser.profile) {                                                                          // 31
    throw new Meteor.Error(400, "Cannot link an identity to another identity.");                      // 32
  }                                                                                                   // 33
                                                                                                      // 34
  if (!!_.findWhere(accountUser.loginIdentities, {id: identityId}) ||                                 // 35
      !!_.findWhere(accountUser.nonloginIdentities, {id: identityId})) {                              // 36
    throw new Meteor.Error(400, "Cannot link an identity that's alread linked to this account.");     // 37
  }                                                                                                   // 38
                                                                                                      // 39
  var identityUser = Meteor.users.findOne({_id: identityId});                                         // 40
                                                                                                      // 41
  if (!identityUser) {                                                                                // 42
    throw new Meteor.Error(404, "No identity found with ID " + identityId);                           // 43
  }                                                                                                   // 44
                                                                                                      // 45
  if (!identityUser.profile) {                                                                        // 46
    throw new Meteor.Error(400, "Cannot link an account to another account");                         // 47
  }                                                                                                   // 48
                                                                                                      // 49
  db.deleteUnusedAccount(backend, identityUser._id);                                                  // 50
  if (Meteor.users.findOne({"loginIdentities.id": identityUser._id})) {                               // 51
    throw new Meteor.Error(403,                                                                       // 52
                           "Cannot link an identity that can already log into another account");      // 53
  }                                                                                                   // 54
                                                                                                      // 55
  var modifier;                                                                                       // 56
  if (accountUser.expires) {                                                                          // 57
    if (Meteor.users.findOne({"nonloginIdentities.id": identityUser._id})) {                          // 58
      throw new Meteor.Error(403, "Cannot create an account for an identity that's " +                // 59
                                  "already linked to another account.");                              // 60
    }                                                                                                 // 61
                                                                                                      // 62
    modifier = {$push: {loginIdentities: {id: identityUser._id}},                                     // 63
                $unset: {expires: 1},                                                                 // 64
                $set: {upgradedFromDemo: Date.now()}};                                                // 65
  } else {                                                                                            // 66
    modifier = {$push: {nonloginIdentities: {id: identityUser._id}}};                                 // 67
  }                                                                                                   // 68
                                                                                                      // 69
  // Make sure not to add the same identity twice.                                                    // 70
  Meteor.users.update({_id: accountUser._id, "nonloginIdentities.id": {$ne: identityUser._id},        // 71
                       "loginIdentities.id": {$ne: identityUser._id}},                                // 72
                      modifier);                                                                      // 73
                                                                                                      // 74
  if (accountUser.expires) {                                                                          // 75
    var demoIdentityId = SandstormDb.getUserIdentityIds(accountUser)[0];                              // 76
    Meteor.users.update({_id: demoIdentityId},                                                        // 77
                        {$unset: {expires: 1},                                                        // 78
                         $set: {upgradedFromDemo: Date.now()}});                                      // 79
                                                                                                      // 80
    // Mark the demo identity as nonlogin. It'd be nicer if the identity started out as nonlogin,     // 81
    // but to get that to work we would need to adjust the account creation and first login logic.    // 82
    Meteor.users.update({_id: accountUser._id,                                                        // 83
                         "loginIdentities.id": demoIdentityId,                                        // 84
                         "nonloginIdentities.id": {$not: {$eq: demoIdentityId}}},                     // 85
                        {$pull: {loginIdentities: {id: demoIdentityId}},                              // 86
                         $push: {nonloginIdentities: {id: demoIdentityId}}});                         // 87
                                                                                                      // 88
  }                                                                                                   // 89
}                                                                                                     // 90
                                                                                                      // 91
Meteor.methods({                                                                                      // 92
  loginWithIdentity: function (accountUserId) {                                                       // 93
    // Logs into the account with ID `accountUserId`. Throws an exception if the current user is      // 94
    // not an identity user listed in the account's `loginIdentities` field. This method is not       // 95
    // intended to be called directly; client-side code should only invoke it through                 // 96
    // `Meteor.loginWithIdentity()`, which additionally maintains the standard Meteor client-side     // 97
    // login state.                                                                                   // 98
                                                                                                      // 99
    check(accountUserId, String);                                                                     // 100
                                                                                                      // 101
    var identityUser = Meteor.user();                                                                 // 102
    if (!identityUser || !identityUser.profile) {                                                     // 103
      throw new Meteor.Error(403, "Must be already logged in as an identity.");                       // 104
    }                                                                                                 // 105
                                                                                                      // 106
    var accountUser = Meteor.users.findOne(accountUserId);                                            // 107
    if (!accountUser) {                                                                               // 108
      throw new Meteor.Error(404, "No such user found: " + accountUserId);                            // 109
    }                                                                                                 // 110
                                                                                                      // 111
    var linkedIdentity = _.findWhere(accountUser.loginIdentities, {id: identityUser._id});            // 112
                                                                                                      // 113
    if (!linkedIdentity) {                                                                            // 114
      throw new Meteor.Error(403, "Current identity is not a login identity for account "             // 115
                             + accountUserId);                                                        // 116
    }                                                                                                 // 117
                                                                                                      // 118
    return Accounts._loginMethod(this, "loginWithIdentity", [accountUserId],                          // 119
                                 "identity", function () { return { userId: accountUserId }; });      // 120
  },                                                                                                  // 121
                                                                                                      // 122
  createAccountForIdentity: function() {                                                              // 123
    // Creates a new account for the currently-logged-in identity.                                    // 124
                                                                                                      // 125
    var user = Meteor.user();                                                                         // 126
    if (!(user && user.profile)) {                                                                    // 127
      throw new Meteor.Error(403, "Must be logged in as an identity in order to create an account.");
    }                                                                                                 // 129
                                                                                                      // 130
    if (Meteor.users.findOne({$or: [{"loginIdentities.id": user._id},                                 // 131
                                    {"nonloginIdentities.id": user._id}]})) {                         // 132
      throw new Meteor.Error(403, "Cannot create an account for an identity that's already " +        // 133
                                  "linked to another account.");                                      // 134
    }                                                                                                 // 135
                                                                                                      // 136
    var newUser = {loginIdentities: [{id: user._id}],                                                 // 137
                   nonloginIdentities: []};                                                           // 138
    if (user.services.dev) {                                                                          // 139
      newUser.signupKey = "devAccounts";                                                              // 140
      if (user.services.dev.isAdmin) {                                                                // 141
        newUser.isAdmin = true;                                                                       // 142
      }                                                                                               // 143
      if (user.services.dev.hasCompletedSignup) {                                                     // 144
        newUser.hasCompletedSignup = true;                                                            // 145
      }                                                                                               // 146
    } else if (user.expires) {                                                                        // 147
      // Demo user.                                                                                   // 148
      newUser.expires = user.expires;                                                                 // 149
      if (!!user.appDemoId) {                                                                         // 150
        newUser.appDemoId = user.appDemoId;                                                           // 151
      }                                                                                               // 152
    }                                                                                                 // 153
    var options = {};                                                                                 // 154
                                                                                                      // 155
    // This will throw an error if the identity has been added as a login identity to some            // 156
    // other account while we were executing the body of this method.                                 // 157
    return Accounts.insertUserDoc(options, newUser);                                                  // 158
  },                                                                                                  // 159
                                                                                                      // 160
  linkIdentityToAccount: function(token) {                                                            // 161
    // Links the identity of the current user to the account that has `token` as a resume token.      // 162
    // If the account is a demo account, makes the account durable and gives the identity login       // 163
    // access to it.                                                                                  // 164
                                                                                                      // 165
    check(token, String);                                                                             // 166
                                                                                                      // 167
    if (!this.userId) {                                                                               // 168
      throw new Meteor.Error(403, "Cannot link to account if not logged in.");                        // 169
    }                                                                                                 // 170
    var hashed = Accounts._hashLoginToken(token);                                                     // 171
    var accountUser = Meteor.users.findOne({"services.resume.loginTokens.hashedToken": hashed});      // 172
                                                                                                      // 173
    linkIdentityToAccountInternal(this.connection.sandstormDb, this.connection.sandstormBackend,      // 174
                                  this.userId, accountUser._id);                                      // 175
  },                                                                                                  // 176
                                                                                                      // 177
  unlinkIdentity: function (accountUserId, identityId) {                                              // 178
    // Unlinks the identity with ID `identityId` from the account with ID `accountUserId`.            // 179
                                                                                                      // 180
    check(identityId, String);                                                                        // 181
    check(accountUserId, String);                                                                     // 182
                                                                                                      // 183
    if (!this.userId) {                                                                               // 184
      throw new Meteor.Error(403, "Not logged in.");                                                  // 185
    }                                                                                                 // 186
    if (!this.connection.sandstormDb.userHasIdentity(this.userId, identityId)) {                      // 187
      throw new Meteor.Error(403, "Current user does not own identity " + identityId);                // 188
    }                                                                                                 // 189
                                                                                                      // 190
    var identityUser = Meteor.users.findOne({_id: identityId});                                       // 191
    Meteor.users.update({_id: accountUserId},                                                         // 192
                        {$pull: {nonloginIdentities: {id: identityId},                                // 193
                                 loginIdentities: {id: identityId}}});                                // 194
  },                                                                                                  // 195
                                                                                                      // 196
  setIdentityAllowsLogin: function(identityId, allowLogin) {                                          // 197
    // Sets whether the current account allows the identity with ID `identityId` to log in.           // 198
                                                                                                      // 199
    check(identityId, String);                                                                        // 200
    check(allowLogin, Boolean);                                                                       // 201
    if (!this.userId) {                                                                               // 202
      throw new Meteor.Error(403, "Not logged in.");                                                  // 203
    }                                                                                                 // 204
    if (!this.connection.sandstormDb.userHasIdentity(this.userId, identityId)) {                      // 205
      throw new Meteor.Error(403, "Current user does not own identity " + identityId);                // 206
    }                                                                                                 // 207
                                                                                                      // 208
    if (allowLogin) {                                                                                 // 209
      Meteor.users.update({_id: this.userId,                                                          // 210
                           "nonloginIdentities.id": identityId,                                       // 211
                           "loginIdentities.id": {$not: {$eq: identityId}}},                          // 212
                          {$pull: {nonloginIdentities: {id: identityId}},                             // 213
                           $push: {loginIdentities: {id: identityId}}});                              // 214
    } else {                                                                                          // 215
      Meteor.users.update({_id: this.userId,                                                          // 216
                           "loginIdentities.id": identityId,                                          // 217
                           "nonloginIdentities.id": {$not: {$eq: identityId}}},                       // 218
                          {$pull: {loginIdentities: {id: identityId}},                                // 219
                           $push: {nonloginIdentities: {id: identityId}}});                           // 220
    }                                                                                                 // 221
  },                                                                                                  // 222
                                                                                                      // 223
  logoutIdentitiesOfCurrentAccount: function() {                                                      // 224
    // Logs out all identities that are allowed to log in to the current account.                     // 225
    var user = Meteor.user();                                                                         // 226
    if (user && user.loginIdentities) {                                                               // 227
      user.loginIdentities.forEach(function(identity) {                                               // 228
        Meteor.users.update({_id: identity.id}, {$set: {"services.resume.loginTokens": []}});         // 229
      });                                                                                             // 230
    }                                                                                                 // 231
  }                                                                                                   // 232
});                                                                                                   // 233
                                                                                                      // 234
Accounts.linkIdentityToAccount = function (db, backend, identityId, accountId) {                      // 235
  // Links the identity to the account. If the account is a demo account, makes it durable and        // 236
  // gives the identity login access to it.                                                           // 237
  check(db, SandstormDb);                                                                             // 238
  check(backend, SandstormBackend);                                                                   // 239
  check(identityId, String);                                                                          // 240
  check(accountId, String);                                                                           // 241
  linkIdentityToAccountInternal(db, backend, identityId, accountId);                                  // 242
}                                                                                                     // 243
                                                                                                      // 244
Meteor.publish("accountsOfIdentity", function (identityId) {                                          // 245
  check(identityId, String);                                                                          // 246
  if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return;                             // 247
                                                                                                      // 248
  // We maintain a map from identity IDs to live query handles that track profile changes.            // 249
  var loginIdentities = {};                                                                           // 250
                                                                                                      // 251
  var self = this;                                                                                    // 252
  function addIdentitiesOfAccount(account) {                                                          // 253
    account.loginIdentities.forEach(function(identity) {                                              // 254
      if (!(identity.id in loginIdentities)) {                                                        // 255
        var user = Meteor.users.findOne({_id: identity.id});                                          // 256
        if (user) {                                                                                   // 257
          SandstormDb.fillInProfileDefaults(user);                                                    // 258
          SandstormDb.fillInIntrinsicName(user);                                                      // 259
          var filteredUser = _.pick(user, "_id", "profile");                                          // 260
          filteredUser.loginAccountId = account._id;                                                  // 261
          filteredUser.sourceIdentityId = identityId;                                                 // 262
          self.added("loginIdentitiesOfLinkedAccounts", user._id, filteredUser);                      // 263
        }                                                                                             // 264
        loginIdentities[identity.id] =                                                                // 265
          Meteor.users.find({_id: identity.id}, {fields: {profile: 1}}).observeChanges({              // 266
            changed: function (id, fields) {                                                          // 267
              self.changed("loginIdentitiesOfLinkedAccounts", id, fields);                            // 268
            }                                                                                         // 269
          });                                                                                         // 270
      }                                                                                               // 271
    });                                                                                               // 272
  }                                                                                                   // 273
  var cursor = Meteor.users.find({$or: [{"loginIdentities.id": identityId},                           // 274
                                        {"nonloginIdentities.id": identityId}]});                     // 275
                                                                                                      // 276
  var handle = cursor.observe({                                                                       // 277
    added: function (account) {                                                                       // 278
      addIdentitiesOfAccount(account);                                                                // 279
    },                                                                                                // 280
    changed: function (newAccount, oldAccount) {                                                      // 281
      addIdentitiesOfAccount(newAccount);                                                             // 282
    },                                                                                                // 283
    removed: function (account) {                                                                     // 284
      account.loginIdentities.forEach(function(identity) {                                            // 285
        if (identity.id in loginIdentities) {                                                         // 286
          self.removed("loginIdentitiesOfLinkedAccounts", identity.id);                               // 287
          loginIdentities[identity.id].stop();                                                        // 288
          delete loginIdentities[identity.id];                                                        // 289
        }                                                                                             // 290
      });                                                                                             // 291
    },                                                                                                // 292
  });                                                                                                 // 293
  this.ready();                                                                                       // 294
                                                                                                      // 295
  this.onStop(function() {                                                                            // 296
    handle.stop();                                                                                    // 297
    Object.keys(loginIdentities).forEach(function(identityId) {                                       // 298
      loginIdentities[identityId].stop();                                                             // 299
      delete loginIdentities[identityId];                                                             // 300
    });                                                                                               // 301
  });                                                                                                 // 302
});                                                                                                   // 303
                                                                                                      // 304
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['accounts-identity'] = {};

})();

//# sourceMappingURL=accounts-identity.js.map
