(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var Packages, DevPackages, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, AppIndex, KeybaseProfiles, isDemoUser, isSignedUp, isSignedUpOrDemo, successfulReferralsCount, getUserQuota, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, roleAssignmentPattern, SandstormDb, i, appName, addStaticAsset, verifiedEmails, unverifiedEmails, userPictureUrl, fetchPicture;

(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/db.js                                                                               //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
// This file defines the database schema.                                                                    // 17
                                                                                                             // 18
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every               // 19
// query it makes, so you can see if it's doing queries too often, etc.                                      // 20
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                                      // 21
  var oldFind = Mongo.Collection.prototype.find;                                                             // 22
  Mongo.Collection.prototype.find = function () {                                                            // 23
    console.log(this._prefix, arguments);                                                                    // 24
    return oldFind.apply(this, arguments);                                                                   // 25
  }                                                                                                          // 26
}                                                                                                            // 27
                                                                                                             // 28
// Users = new Mongo.Collection("users");                                                                    // 29
// The users collection is special and can be accessed through `Meteor.users`.                               // 30
// See https://docs.meteor.com/#/full/meteor_users.                                                          // 31
//                                                                                                           // 32
// There are two distinct types of entries in the users collection: identities and accounts. An              // 33
// identity contains personal profile information and typically includes some intrinsic method for           // 34
// authenticating as the owner of that information.                                                          // 35
//                                                                                                           // 36
// An account is an owner of app actions, grains, contacts, notifications, and payment info.                 // 37
// Each account can have multiple identities linked to it. To log in as an account you must first            // 38
// authenticate as one of its linked identities.                                                             // 39
//                                                                                                           // 40
// Every user contains the following fields:                                                                 // 41
//   _id: Unique string ID. For accounts, this is random. For identities, this is the globally               // 42
//        stable SHA-256 ID of this identity, hex-encoded.                                                   // 43
//   createdAt: Date when this entry was added to the collection.                                            // 44
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.                      // 45
//   services: Object containing login data used by Meteor authentication services.                          // 46
//   expires: Date when this user should be deleted. Only present for demo users.                            // 47
//   upgradedFromDemo: If present, the date when this user was upgraded from being a demo user.              // 48
//                     TODO(cleanup): Unlike other dates in our database, this is stored as a number         // 49
//                     rather than as a Date object. We should fix that.                                     // 50
//   appDemoId: If present and non-null, then the user is a demo user who arrived via an /appdemo/           // 51
//              link. This field contains the app ID of the app that the user started out demoing.           // 52
//              Unlike the `expires` field, this field is not cleared when the user upgrades from            // 53
//              being a demo user.                                                                           // 54
//                                                                                                           // 55
// Identity users additionally contain the following fields:                                                 // 56
//   profile: Object containing the data that will be shared with users and grains that come into            // 57
//            contact with this identity. Includes the following fields:                                     // 58
//       service: String containing the name of this identity's authentication method.                       // 59
//       name: String containing the chosen display name of the identity.                                    // 60
//       handle: String containing the identity's preferred handle.                                          // 61
//       picture: _id into the StaticAssets table for the identity's picture. If not present,                // 62
//                an identicon will be used.                                                                 // 63
//       pronoun: One of "male", "female", "neutral", or "robot".                                            // 64
//   unverifiedEmail: If present, a string containing an email address specified by the user.                // 65
//   referredBy: ID of the Account that referred this Identity.                                              // 66
//                                                                                                           // 67
// Account users additionally contain the following fields:                                                  // 68
//   loginIdentities: Array of identity objects, each of which may include the following fields.             // 69
//       id: The globally-stable SHA-256 ID of this identity, hex-encoded.                                   // 70
//   nonloginIdentities: Array of identity objects, of the same form as `loginIdentities`. We use            // 71
//                       a separate array here so that we can use a Mongo index to enforce the               // 72
//                       invariant that an identity only be a login identity for a single account.           // 73
//   primaryEmail: String containing this account's primary email address. Must be a verified adress         // 74
//                 of one of this account's linked identities. Call SandstormDb.getUserEmails()              // 75
//                 to do this checking automatically.                                                        // 76
//   isAdmin: Boolean indicating whether this account is allowed to access the Sandstorm admin panel.        // 77
//   signupKey: If this is an invited user, then this field contains their signup key.                       // 78
//   signupNote: If the user was invited through a link, then this field contains the note that the          // 79
//               inviter admin attached to the key.                                                          // 80
//   signupEmail: If the user was invited by email, then this field contains the email address that          // 81
//                the invite was sent to.                                                                    // 82
//   hasCompletedSignup: True if this account has confirmed its profile and agreed to this server's          // 83
//                       terms of service.                                                                   // 84
//   plan: _id of an entry in the Plans table which determines the user's qutoa.                             // 85
//   storageUsage: Number of bytes this user is currently storing.                                           // 86
//   payments: Object defined by payments module, if loaded.                                                 // 87
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                       // 88
//   referredByComplete: ID of the Account that referred this Account. If this is set, we                    // 89
//                        stop writing new referredBy values onto Identities for this account.               // 90
//   referredCompleteDate: The Date at which the completed referral occurred.                                // 91
//   referredIdentityIds: List of Identity IDs that this Account has referred. This is used for              // 92
//                        reliably determining which Identity's names are safe to display.                   // 93
//   experiments: Object where each field is an experiment that the user is in, and each value               // 94
//           is the parameters for that experiment. Typically, the value simply names which                  // 95
//           experiment group which the user is in, where "control" is one group. If an experiment           // 96
//           is not listed, then the user should not be considered at all for the purpose of that            // 97
//           experiment. Each experiment may define a point in time where users not already in the           // 98
//           experiment may be added to it and assigned to a group (for example, at user creation            // 99
//           time). Current experiments:                                                                     // 100
//       firstTimeBillingPrompt: Value is "control" or "test". Users are assigned to groups at               // 101
//               account creation on servers where billing is enabled (i.e. Oasis). Users in the             // 102
//               test group will see a plan selection dialog and asked to make an explitic choice            // 103
//               (possibly "free") before they can create grains (but not when opening someone               // 104
//               else's shared grain). The goal of the experiment is to determine whether this               // 105
//               prompt scares users away -- and also whether it increases paid signups.                     // 106
//   stashedOldUser: A complete copy of this user from before the accounts/identities migration.             // 107
//                   TODO(cleanup): Delete this field once we're sure it's safe to do so.                    // 108
                                                                                                             // 109
Packages = new Mongo.Collection("packages");                                                                 // 110
// Packages which are installed or downloading.                                                              // 111
//                                                                                                           // 112
// Each contains:                                                                                            // 113
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                          // 114
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"         // 115
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),                      // 116
//       >1 = download byte count.                                                                           // 117
//   error:  If status is "failed", error message string.                                                    // 118
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.                // 119
//   appId:  If status is "ready", the application ID string.  Packages representing different               // 120
//       versions of the same app have the same appId.  The spk tool defines the app ID format               // 121
//       and can cryptographically verify that a package belongs to a particular app ID.                     // 122
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package              // 123
//       collector should at some point check whether there are any other references and, if not,            // 124
//       delete the package.                                                                                 // 125
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.               // 126
//   isAutoUpdated: This package was downloaded as part of an auto-update. We shouldn't clean it up          // 127
//     even if it has no users.                                                                              // 128
//   authorPgpKeyFingerprint: Verified PGP key fingerprint (SHA-1, hex, all-caps) of the app                 // 129
//     packager.                                                                                             // 130
                                                                                                             // 131
DevPackages = new Mongo.Collection("devpackages");                                                           // 132
// List of packages currently made available via the dev tools running on the local machine.                 // 133
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool           // 134
// on the local machine to publish an under-development app to this server. That should only ever            // 135
// happen on developers' desktop machines.                                                                   // 136
//                                                                                                           // 137
// While a dev package is published, it automatically appears as installed by every user of the              // 138
// server, and it overrides all packages with the same application ID. If any instances of those             // 139
// packages are currently open, they are killed and reset on publish.                                        // 140
//                                                                                                           // 141
// When the dev tool disconnects, the package is automatically unpublished, and any open instances           // 142
// are again killed and refreshed.                                                                           // 143
//                                                                                                           // 144
// Each contains:                                                                                            // 145
//   _id:  The package ID string (as with Packages._id).                                                     // 146
//   appId: The app ID this package is intended to override (as with Packages.appId).                        // 147
//   timestamp:  Time when the package was last updated. If this changes while the package is                // 148
//     published, all running instances are reset. This is used e.g. to reset the app each time              // 149
//     changes are made to the source code.                                                                  // 150
//   manifest:  The app's manifest, as with Packages.manifest.                                               // 151
                                                                                                             // 152
UserActions = new Mongo.Collection("userActions");                                                           // 153
// List of actions that each user has installed which create new grains.  Each app may install               // 154
// some number of actions (usually, one).                                                                    // 155
//                                                                                                           // 156
// Each contains:                                                                                            // 157
//   _id:  random                                                                                            // 158
//   userId:  User who has installed this action.                                                            // 159
//   packageId:  Package used to run this action.                                                            // 160
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 161
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                       // 162
//       that clients can access it without subscribing to the Packages collection.                          // 163
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 164
//       searchability.                                                                                      // 165
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"                     // 166
//   title: JSON-encoded LocalizedText title for this action, e.g.                                           // 167
//       `{defaultText: "New Spreadsheet"}`.                                                                 // 168
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.              // 169
//   command:  Manifest.Command to run this action (see package.capnp).                                      // 170
                                                                                                             // 171
Grains = new Mongo.Collection("grains");                                                                     // 172
// Grains belonging to users.                                                                                // 173
//                                                                                                           // 174
// Each contains:                                                                                            // 175
//   _id:  random                                                                                            // 176
//   packageId:  _id of the package of which this grain is an instance.                                      // 177
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 178
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 179
//       searchability.                                                                                      // 180
//   userId: The _id of the account that owns this grain.                                                    // 181
//   identityId: The identity with which the owning account prefers to open this grain.                      // 182
//   title:  Human-readable string title, as chosen by the user.                                             // 183
//   lastUsed:  Date when the grain was last used by a user.                                                 // 184
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                          // 185
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent          // 186
//                   time a session to this grain was opened.                                                // 187
//                                                                                                           // 188
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and           // 189
// web publishing functionality without powerbox support; they will be replaced once the powerbox            // 190
// is implemented.                                                                                           // 191
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and           // 192
//       web publishing. This field is initialized when first requested by the app.                          // 193
                                                                                                             // 194
RoleAssignments = new Mongo.Collection("roleAssignments");                                                   // 195
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges          // 196
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens                   // 197
// collection.                                                                                               // 198
                                                                                                             // 199
Contacts = new Mongo.Collection("contacts");                                                                 // 200
// Edges in the social graph.                                                                                // 201
//                                                                                                           // 202
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob               // 203
// will show up in her user-picker UI for actions like share-by-identity.                                    // 204
//                                                                                                           // 205
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of              // 206
// Bob's.                                                                                                    // 207
//                                                                                                           // 208
// Each contains:                                                                                            // 209
//   _id: random                                                                                             // 210
//   ownerId: The accountId of the user account who owns this contact.                                       // 211
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies              // 212
//            the contact to the owner.                                                                      // 213
//   created: Date when this contact was created.                                                            // 214
//   identityId: The `_id` of the user whose contact info this contains.                                     // 215
                                                                                                             // 216
Sessions = new Mongo.Collection("sessions");                                                                 // 217
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.          // 218
//                                                                                                           // 219
// Each contains:                                                                                            // 220
//   _id:  random                                                                                            // 221
//   grainId:  _id of the grain to which this session is connected.                                          // 222
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the           // 223
//       '*' in WILDCARD_HOST.                                                                               // 224
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some              // 225
//       period.                                                                                             // 226
//   userId:  User ID of the user who owns this session.                                                     // 227
//   identityId:  Identity ID of the user who owns this session.                                             // 228
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens            // 229
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),               // 230
//       anonymous users can get access without an API token and so neither userId nor hashedToken           // 231
//       are present.                                                                                        // 232
//   powerboxView: If present, this is a view that should be presented as part of a powerbox                 // 233
//       interaction.                                                                                        // 234
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                         // 235
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated           // 236
//       until newSession is called on the UiView.                                                           // 237
//   permissions: The permissions for the current identity on this UiView. This isn't populated              // 238
//       until newSession is called on the UiView.                                                           // 239
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first         // 240
//       request                                                                                             // 241
                                                                                                             // 242
SignupKeys = new Mongo.Collection("signupKeys");                                                             // 243
// Invite keys which may be used by users to get access to Sandstorm.                                        // 244
//                                                                                                           // 245
// Each contains:                                                                                            // 246
//   _id:  random                                                                                            // 247
//   used:  Boolean indicating whether this key has already been consumed.                                   // 248
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.                // 249
//   email: If this key was sent as an email invite, the email address to which it was sent.                 // 250
                                                                                                             // 251
ActivityStats = new Mongo.Collection("activityStats");                                                       // 252
// Contains usage statistics taken on a regular interval. Each entry is a data point.                        // 253
//                                                                                                           // 254
// Each contains:                                                                                            // 255
//   timestamp: Date when measurements were taken.                                                           // 256
//   daily: Contains stats counts pertaining to the last day before the sample time.                         // 257
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.                 // 258
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.                 // 259
//                                                                                                           // 260
// Each of daily, weekly, and monthly contains:                                                              // 261
//   activeUsers: The number of unique users who have used a grain on the server in the time                 // 262
//       interval. Only counts logged-in users.                                                              // 263
//   demoUsers: Demo users.                                                                                  // 264
//   appDemoUsers: Users that came in through "app demo".                                                    // 265
//   activeGrains: The number of unique grains that have been used in the time interval.                     // 266
//   apps: An object indexed by app ID recording, for each app:                                              // 267
//       owners: Number of unique owners of this app (counting only grains that still exist).                // 268
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting          // 269
//         only grains that still exist).                                                                    // 270
//       grains: Number of active grains of this app (that still exist).                                     // 271
//       deleted: Number of non-demo grains of this app that were deleted.                                   // 272
//       demoed: Number of demo grains created and expired.                                                  // 273
//       appDemoUsers: Number of app demos initiated with this app.                                          // 274
                                                                                                             // 275
DeleteStats = new Mongo.Collection("deleteStats");                                                           // 276
// Contains records of objects that were deleted, for stat-keeping purposes.                                 // 277
//                                                                                                           // 278
// Each contains:                                                                                            // 279
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                                   // 280
//   lastActive: Date of the user's or grain's last activity.                                                // 281
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they           // 282
//     arrived to demo. For others, undefined.                                                               // 283
                                                                                                             // 284
FileTokens = new Mongo.Collection("fileTokens");                                                             // 285
// Tokens corresponding to backup files that are currently stored on the server. A user receives             // 286
// a token when they create a backup file (either by uploading it, or by backing up one of their             // 287
// grains) and may use the token to read the file (either to download it, or to restore a new                // 288
// grain from it).                                                                                           // 289
//                                                                                                           // 290
// Each contains:                                                                                            // 291
//   _id:       The unguessable token string.                                                                // 292
//   name:      Suggested filename.                                                                          // 293
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.              // 294
                                                                                                             // 295
ApiTokens = new Mongo.Collection("apiTokens");                                                               // 296
// Access tokens for APIs exported by apps.                                                                  // 297
//                                                                                                           // 298
// Originally API tokens were only used by external users through the HTTP API endpoint. However,            // 299
// now they are also used to implement SturdyRefs, not just held by external users, but also when            // 300
// an app holds a SturdyRef to another app within the same server. See the various `save()`,                 // 301
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and               // 302
// `MainView`) -- the fields of type `Data` are API tokens.                                                  // 303
//                                                                                                           // 304
// Each contains:                                                                                            // 305
//   _id:       A SHA-256 hash of the token, base64-encoded.                                                 // 306
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)            // 307
//   identityId: For UiView capabilities, this is the identity for which the view is attenuated.             // 308
//              That is, the UiView's newSession() method will intersect the requested permissions           // 309
//              with this identity's permissions before forwarding on to the underlying app. If              // 310
//              `identityId` is not present, then no identity attenuation is applied, i.e. this is           // 311
//              a raw UiView as implemented by the app. (The `roleAssignment` field, below, may              // 312
//              still apply. For non-UiView capabilities, `identityId` is never present. Note that           // 313
//              this is NOT the identity against which the `requiredPermissions` parameter of                // 314
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerIdentity`.         // 315
//   accountId: For tokens where `identityId` is set, the `_id` (in the Users table) of the account          // 316
//              that created the token.                                                                      // 317
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded               // 318
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These          // 319
//              permissions will be intersected with those held by `identityId` when the view is             // 320
//              opened.                                                                                      // 321
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as          // 322
//              anonymous rather than as directly associated with `identityId`. This has no effect           // 323
//              on the permissions granted.                                                                  // 324
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by            // 325
//              the app or its supervisor (whereas without this it strictly represents UiView).              // 326
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).            // 327
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is                   // 328
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.                     // 329
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,          // 330
//              not a particular grain. (`grainId` and `identityId` are not set.) This is an object          // 331
//              containing exactly one of the following fields:                                              // 332
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                          // 333
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the                 // 334
//                           `Notifications` collection.                                                     // 335
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this            // 336
//                  will be moved out of the frontend and into the backend, but we'll migrate the            // 337
//                  database when that happens. This field contains the boolean true to signify that         // 338
//                  it has been set.                                                                         // 339
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                           // 340
//   parentToken: If present, then this token represents exactly the capability represented by               // 341
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated          // 342
//              by `roleAssignment` (if present). To facilitate permissions computations, if the             // 343
//              capability is a UiView, then `grainId` is set to the backing grain, `identityId`             // 344
//              is set to the identity that shared the view, and `accountId` is set to the account           // 345
//              that shared the view. Neither `objectId` nor `frontendRef` is present when                   // 346
//              `parentToken` is present.                                                                    // 347
//   petname:   Human-readable label for this access token, useful for identifying tokens for                // 348
//              revocation. This should be displayed when visualizing incoming capabilities to               // 349
//              the grain identified by `grainId`.                                                           // 350
//   created:   Date when this token was created.                                                            // 351
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may          // 352
//              become un-revoked in the future.                                                             // 353
//   expires:   Optional expiration Date. If undefined, the token does not expire.                           // 354
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)                // 355
//              as passed to the `save()` call that created this token. If not present, treat                // 356
//              as `webkey` (the default for `ApiTokenOwner`).                                               // 357
//   expiresIfUnused:                                                                                        // 358
//              Optional Date after which the token, if it has not been used yet, expires.                   // 359
//              This field should be cleared on a token's first use.                                         // 360
//   requirements: List of conditions which must hold for this token to be considered valid.                 // 361
//              Semantically, this list specifies the powers which were *used* to originally                 // 362
//              create the token. If any condition in the list becomes untrue, then the token must           // 363
//              be considered revoked, and all live refs and sturdy refs obtained transitively               // 364
//              through it must also become revoked. Each item is the JSON serialization of the              // 365
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                               // 366
//                                                                                                           // 367
// It is important to note that a token's owner and provider are independent from each other. To             // 368
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:          // 369
//                                                                                                           // 370
// struct ApiToken {                                                                                         // 371
//   owner :ApiTokenOwner;                                                                                   // 372
//   provider :union {                                                                                       // 373
//     grain :group {                                                                                        // 374
//       grainId :Text;                                                                                      // 375
//       union {                                                                                             // 376
//         uiView :group {                                                                                   // 377
//           identityId :Text;                                                                               // 378
//           roleAssignment :RoleAssignment;                                                                 // 379
//           forSharing :Bool;                                                                               // 380
//         }                                                                                                 // 381
//         objectId :SupervisorObjectId;                                                                     // 382
//       }                                                                                                   // 383
//     }                                                                                                     // 384
//     frontendRef :union {                                                                                  // 385
//        notificationHandle :Text;                                                                          // 386
//        ipNetwork :Bool;                                                                                   // 387
//        ipInterface :Bool;                                                                                 // 388
//     }                                                                                                     // 389
//     child :group {                                                                                        // 390
//       parentToken :Text;                                                                                  // 391
//       union {                                                                                             // 392
//         uiView :group {                                                                                   // 393
//           grainId :Text;                                                                                  // 394
//           identityId :Text;                                                                               // 395
//           roleAssignment :RoleAssignment = (allAccess = ());                                              // 396
//         }                                                                                                 // 397
//         other :Void;                                                                                      // 398
//       }                                                                                                   // 399
//     }                                                                                                     // 400
//   }                                                                                                       // 401
//   requirements: List(Supervisor.MembraneRequirement);                                                     // 402
//   ...                                                                                                     // 403
// }                                                                                                         // 404
                                                                                                             // 405
Notifications = new Mongo.Collection("notifications");                                                       // 406
// Notifications for a user.                                                                                 // 407
//                                                                                                           // 408
// Each contains:                                                                                            // 409
//   _id:          random                                                                                    // 410
//   grainId:      The grain originating this notification, if any.                                          // 411
//   userId:       The user receiving the notification.                                                      // 412
//   text:         The JSON-ified LocalizedText to display in the notification.                              // 413
//   isUnread:     Boolean indicating if this notification is unread.                                        // 414
//   timestamp:    Date when this notification was last updated                                              // 415
//   ongoing:      If present, this is an ongoing notification, and this field contains an                   // 416
//                 ApiToken referencing the `OngoingNotification` capability.                                // 417
//   admin:        If present, this is a notification intended for an admin.                                 // 418
//     action:     If present, this is a (string) link that the notification should direct the               // 419
//                 admin to.                                                                                 // 420
//     type:       The type of notification (currently only "reportStats").                                  // 421
//   appUpdates:   If present, this is an app update notification. It is an object with the appIds           // 422
//                 as keys.                                                                                  // 423
//     $appId:     The appId that has an outstanding update.                                                 // 424
//       packageId: The packageId that it will update to.                                                    // 425
//       name: The name of the app. (appTitle from package.manifest)                                         // 426
//       version: The app's version number. (appVersion from package.manifest)                               // 427
//       marketingVersion: String marketing version of this app. (appMarketingVersion from package.manifest)
                                                                                                             // 429
StatsTokens = new Mongo.Collection("statsTokens");                                                           // 430
// Access tokens for the Stats collection                                                                    // 431
//                                                                                                           // 432
// These tokens are used for accessing the ActivityStats collection remotely                                 // 433
// (ie. from a dashboard webapp)                                                                             // 434
//                                                                                                           // 435
// Each contains:                                                                                            // 436
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                        // 437
                                                                                                             // 438
Misc = new Mongo.Collection("misc");                                                                         // 439
// Miscellaneous configuration and other settings                                                            // 440
//                                                                                                           // 441
// This table is currently only used for persisting BASE_URL from one session to the next,                   // 442
// but in general any miscellaneous settings should go in here                                               // 443
//                                                                                                           // 444
// Each contains:                                                                                            // 445
//   _id:       The name of the setting. eg. "BASE_URL"                                                      // 446
//   value:     The value of the setting.                                                                    // 447
                                                                                                             // 448
Settings = new Mongo.Collection("settings");                                                                 // 449
// Settings for this Sandstorm instance go here. They are configured through the adminSettings               // 450
// route. This collection differs from misc in that this collection is completely user controlled.           // 451
//                                                                                                           // 452
// Each contains:                                                                                            // 453
//   _id:       The name of the setting. eg. "MAIL_URL"                                                      // 454
//   value:     The value of the setting.                                                                    // 455
//   potentially other fields that are unique to the setting                                                 // 456
                                                                                                             // 457
Migrations = new Mongo.Collection("migrations");                                                             // 458
// This table tracks which migrations we have applied to this instance.                                      // 459
// It contains a single entry:                                                                               // 460
//   _id:       "migrations_applied"                                                                         // 461
//   value:     The number of migrations this instance has successfully completed.                           // 462
                                                                                                             // 463
StaticAssets = new Mongo.Collection("staticAssets");                                                         // 464
// Collection of static assets served up from the Sandstorm server's "static" host. We only                  // 465
// support relatively small assets: under 1MB each.                                                          // 466
//                                                                                                           // 467
// Each contains:                                                                                            // 468
//   _id:       Random ID; will be used in the URL.                                                          // 469
//   hash:      A base64-encoded SHA-256 hash of the data, used to de-dupe.                                  // 470
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                                    // 471
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                          // 472
//   content:   The asset content (byte buffer).                                                             // 473
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't          // 474
//       have transactions, this needs to bias towards over-counting; a backup GC could be used              // 475
//       to catch leaked assets, although it's probably not a big deal in practice.                          // 476
                                                                                                             // 477
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                               // 478
// Collection of tokens representing a single-use permission to upload an asset, such as a new               // 479
// profile picture.                                                                                          // 480
//                                                                                                           // 481
// Each contains:                                                                                            // 482
//   _id:       Random ID.                                                                                   // 483
//   purpose:   Contains one of the following, indicating how the asset is to be used:                       // 484
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:                // 485
//           userId: User whose picture shall be replaced.                                                   // 486
//           identityId: Which of the user's identities shall be updated.                                    // 487
//   expires:   Time when this token will go away if unused.                                                 // 488
                                                                                                             // 489
Plans = new Mongo.Collection("plans");                                                                       // 490
// Subscription plans, which determine quota.                                                                // 491
//                                                                                                           // 492
// Each contains:                                                                                            // 493
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...                      // 494
//   storage: Number of bytes this user is allowed to store.                                                 // 495
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                                // 496
//   computeLabel: Label to display to the user describing this plan's compute units.                        // 497
//   grains: Total number of grains this user can create (often `Infinity`).                                 // 498
//   price: Price per month in US cents.                                                                     // 499
//   hidden: If true, a user cannot switch to this plan, but some users may be on it and are                 // 500
//       allowed to switch away.                                                                             // 501
//   title: Title from display purposes. If missing, default to capitalizing _id.                            // 502
                                                                                                             // 503
AppIndex = new Mongo.Collection("appIndex");                                                                 // 504
// A mirror of the data from the App Market index                                                            // 505
//                                                                                                           // 506
// Each contains:                                                                                            // 507
//   _id: the appId of the app                                                                               // 508
//  The rest of the fields are defined in src/sandstorm/app-index/app-index.capnp:AppIndexForMarket          // 509
                                                                                                             // 510
KeybaseProfiles = new Mongo.Collection("keybaseProfiles");                                                   // 511
// Cache of Keybase profile information. The profile for a user is re-fetched every time a package           // 512
// by that user is installed, as well as if the keybase profile is requested and not already                 // 513
// present for some reason.                                                                                  // 514
//                                                                                                           // 515
// Each contains:                                                                                            // 516
//   _id: PGP key fingerprint (SHA-1, hex, all-caps)                                                         // 517
//   displayName: Display name from Keybase. (NOT VERIFIED AT ALL.)                                          // 518
//   handle: Keybase handle.                                                                                 // 519
//   proofs: The "proofs_summary.all" array from the Keybase lookup. See the non-existent Keybase            // 520
//     docs for details. We also add a boolean "status" field to each proof indicating whether               // 521
//     we have directly verified the proof ourselves. Its values may be "unverified" (Keybase                // 522
//     returned this but we haven't checked it directly), "verified" (we verified the proof and it           // 523
//     is valid), "invalid" (we checked the proof and it was definitely bogus), or "checking" (the           // 524
//     server is currently actively checking this proof). Note that if a check fails due to network          // 525
//     errors, the status goes back to "unverified".                                                         // 526
//                                                                                                           // 527
//     WARNING: Currently verification is NOT IMPLEMENTED, so all proofs will be "unverified"                // 528
//       for now and we just trust Keybase.                                                                  // 529
                                                                                                             // 530
if (Meteor.isServer) {                                                                                       // 531
  Meteor.publish("credentials", function () {                                                                // 532
    // Data needed for isSignedUp() and isAdmin() to work.                                                   // 533
                                                                                                             // 534
    if (this.userId) {                                                                                       // 535
      return [                                                                                               // 536
        Meteor.users.find({_id: this.userId},                                                                // 537
            {fields: {signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,                                 // 538
                      plan: 1, hasCompletedSignup: 1, experiments: 1}}),                                     // 539
        Plans.find()                                                                                         // 540
      ];                                                                                                     // 541
    } else {                                                                                                 // 542
      return [];                                                                                             // 543
    }                                                                                                        // 544
  });                                                                                                        // 545
}                                                                                                            // 546
                                                                                                             // 547
isDemoUser = function() {                                                                                    // 548
  // Returns true if this is a demo user.                                                                    // 549
                                                                                                             // 550
  var user = Meteor.user();                                                                                  // 551
  if (user && user.expires) {                                                                                // 552
    return true;                                                                                             // 553
  } else {                                                                                                   // 554
    return false;                                                                                            // 555
  }                                                                                                          // 556
}                                                                                                            // 557
                                                                                                             // 558
isSignedUp = function() {                                                                                    // 559
  // Returns true if the user has presented an invite key.                                                   // 560
                                                                                                             // 561
  var user = Meteor.user();                                                                                  // 562
                                                                                                             // 563
  if (!user) return false;  // not signed in                                                                 // 564
                                                                                                             // 565
  if (!user.loginIdentities) return false;  // not an account                                                // 566
                                                                                                             // 567
  if (user.expires) return false;  // demo user.                                                             // 568
                                                                                                             // 569
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 570
                                                                                                             // 571
  if (user.signupKey) return true;  // user is invited                                                       // 572
                                                                                                             // 573
  return false;                                                                                              // 574
}                                                                                                            // 575
                                                                                                             // 576
isSignedUpOrDemo = function () {                                                                             // 577
  var user = Meteor.user();                                                                                  // 578
                                                                                                             // 579
  if (!user) return false;  // not signed in                                                                 // 580
                                                                                                             // 581
  if (!user.loginIdentities) return false;  // not an account                                                // 582
                                                                                                             // 583
  if (user.expires) return true;  // demo user.                                                              // 584
                                                                                                             // 585
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 586
                                                                                                             // 587
  if (user.signupKey) return true;  // user is invited                                                       // 588
                                                                                                             // 589
  return false;                                                                                              // 590
}                                                                                                            // 591
                                                                                                             // 592
var calculateReferralBonus = function(accountId, plan) {                                                     // 593
  // This function returns an object of the form:                                                            // 594
  //                                                                                                         // 595
  // - {grains: 0, storage: 0}                                                                               // 596
  //                                                                                                         // 597
  // which are extra resources this account gets as part of participating in the referral                    // 598
  // program. (Storage is measured in bytes, as usual for plans.)                                            // 599
                                                                                                             // 600
                                                                                                             // 601
  // Authorization note: Only call this if accountId is the current user!                                    // 602
  var isPaid = (plan && plan !== "free");                                                                    // 603
                                                                                                             // 604
  successfulReferralsCount = countReferrals(accountId);                                                      // 605
  if (isPaid) {                                                                                              // 606
    var maxPaidStorageBonus = 30 * 1e9;                                                                      // 607
    return {grains: 0,                                                                                       // 608
            storage: Math.min(                                                                               // 609
              successfulReferralsCount * 2 * 1e9,                                                            // 610
              maxPaidStorageBonus)};                                                                         // 611
  } else {                                                                                                   // 612
    var maxFreeStorageBonus = 2 * 1e9;                                                                       // 613
    var bonus = {                                                                                            // 614
      storage: Math.min(                                                                                     // 615
        successfulReferralsCount * 50 * 1e6,                                                                 // 616
        maxFreeStorageBonus)                                                                                 // 617
    };                                                                                                       // 618
    if (successfulReferralsCount > 0) {                                                                      // 619
      bonus.grains = Infinity;                                                                               // 620
    } else {                                                                                                 // 621
      bonus.grains = 0;                                                                                      // 622
    }                                                                                                        // 623
    return bonus;                                                                                            // 624
  }                                                                                                          // 625
}                                                                                                            // 626
                                                                                                             // 627
var countReferrals = function (accountId) {                                                                  // 628
  var referredIdentityIds = Meteor.users.findOne(                                                            // 629
    {_id: accountId},                                                                                        // 630
    {fields: {referredIdentityIds: 1}}).referredIdentityIds;                                                 // 631
  return (referredIdentityIds && referredIdentityIds.length || 0);                                           // 632
}                                                                                                            // 633
                                                                                                             // 634
getUserQuota = function (user) {                                                                             // 635
  var plan = Plans.findOne(user.plan || "free");                                                             // 636
  var referralBonus = calculateReferralBonus(user._id, plan);                                                // 637
  var userQuota = {};                                                                                        // 638
  userQuota.storage = (plan.storage + referralBonus.storage);                                                // 639
  userQuota.grains = (plan.grains + referralBonus.grains);                                                   // 640
  return userQuota;                                                                                          // 641
}                                                                                                            // 642
                                                                                                             // 643
isUserOverQuota = function (user) {                                                                          // 644
  // Return false if user has quota space remaining, true if it is full. When this returns true,             // 645
  // we will not allow the user to create new grains, though they may be able to open existing ones          // 646
  // which may still increase their storage usage.                                                           // 647
  //                                                                                                         // 648
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 649
                                                                                                             // 650
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 651
                                                                                                             // 652
  var plan = getUserQuota(user);                                                                             // 653
  if (plan.grains < Infinity) {                                                                              // 654
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains}).count();                   // 655
    if (count >= plan.grains) return "outOfGrains";                                                          // 656
  }                                                                                                          // 657
                                                                                                             // 658
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";                   // 659
}                                                                                                            // 660
                                                                                                             // 661
isUserExcessivelyOverQuota = function (user) {                                                               // 662
  // Return true if user is so far over quota that we should prevent their existing grains from              // 663
  // running at all.                                                                                         // 664
  //                                                                                                         // 665
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 666
                                                                                                             // 667
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 668
                                                                                                             // 669
  var quota = getUserQuota(user);                                                                            // 670
                                                                                                             // 671
  // quota.grains = Infinity means unlimited grains. IEEE754 defines Infinity == Infinity.                   // 672
  if (quota.grains < Infinity) {                                                                             // 673
    var count = Grains.find({userId: user._id}, {fields: {}, limit: quota.grains * 2}).count();              // 674
    if (count >= quota.grains * 2) return "outOfGrains";                                                     // 675
  }                                                                                                          // 676
                                                                                                             // 677
  return quota && user.storageUsage && user.storageUsage >= quota.storage * 1.2 && "outOfStorage";           // 678
}                                                                                                            // 679
                                                                                                             // 680
isAdmin = function() {                                                                                       // 681
  // Returns true if the user is the administrator.                                                          // 682
                                                                                                             // 683
  var user = Meteor.user();                                                                                  // 684
  if (user && user.isAdmin) {                                                                                // 685
    return true;                                                                                             // 686
  } else {                                                                                                   // 687
    return false;                                                                                            // 688
  }                                                                                                          // 689
}                                                                                                            // 690
                                                                                                             // 691
isAdminById = function(id) {                                                                                 // 692
  // Returns true if the user's id is the administrator.                                                     // 693
                                                                                                             // 694
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                         // 695
  if (user && user.isAdmin) {                                                                                // 696
    return true;                                                                                             // 697
  } else {                                                                                                   // 698
    return false;                                                                                            // 699
  }                                                                                                          // 700
}                                                                                                            // 701
                                                                                                             // 702
findAdminUserForToken = function (token) {                                                                   // 703
  if (!token.requirements) {                                                                                 // 704
    return;                                                                                                  // 705
  }                                                                                                          // 706
  var requirements = token.requirements.filter(function (requirement) {                                      // 707
    return "userIsAdmin" in requirement;                                                                     // 708
  });                                                                                                        // 709
                                                                                                             // 710
  if (requirements.length > 1) {                                                                             // 711
    return;                                                                                                  // 712
  }                                                                                                          // 713
  if (requirements.length === 0) {                                                                           // 714
    return;                                                                                                  // 715
  }                                                                                                          // 716
  return requirements[0].userIsAdmin;                                                                        // 717
};                                                                                                           // 718
                                                                                                             // 719
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                             // 720
                                                                                                             // 721
if (wildcardHost.length != 2) {                                                                              // 722
  throw new Error("Wildcard host must contain exactly one asterisk.");                                       // 723
}                                                                                                            // 724
                                                                                                             // 725
matchWildcardHost = function(host) {                                                                         // 726
  // See if the hostname is a member of our wildcard. If so, extract the ID.                                 // 727
                                                                                                             // 728
  var prefix = wildcardHost[0];                                                                              // 729
  var suffix = wildcardHost[1];                                                                              // 730
                                                                                                             // 731
  // We remove everything after the first ":" character so that our                                          // 732
  // comparison logic ignores port numbers.                                                                  // 733
  suffix = suffix.split(":")[0];                                                                             // 734
  host = host.split(":")[0];                                                                                 // 735
                                                                                                             // 736
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                                    // 737
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                           // 738
      host.length >= prefix.length + suffix.length) {                                                        // 739
    var id = host.slice(prefix.length, -suffix.length);                                                      // 740
    if (id.match(/^[a-z0-9]*$/)) {                                                                           // 741
      return id;                                                                                             // 742
    }                                                                                                        // 743
  }                                                                                                          // 744
                                                                                                             // 745
  return null;                                                                                               // 746
}                                                                                                            // 747
                                                                                                             // 748
makeWildcardHost = function (id) {                                                                           // 749
  return wildcardHost[0] + id + wildcardHost[1];                                                             // 750
}                                                                                                            // 751
                                                                                                             // 752
if (Meteor.isServer) {                                                                                       // 753
  var Url = Npm.require("url");                                                                              // 754
  getWildcardOrigin = function () {                                                                          // 755
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're              // 756
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),            // 757
    // an astrisk is only allowed as the first character and must be followed by a period. So we need        // 758
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can           // 759
    // really do. We also add the protocol to the front (again, that's what CSP wants).                      // 760
                                                                                                             // 761
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.         // 762
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                                 // 763
                                                                                                             // 764
    var dotPos = wildcardHost[1].indexOf(".");                                                               // 765
    if (dotPos < 0) {                                                                                        // 766
      return protocol + "//*";                                                                               // 767
    } else {                                                                                                 // 768
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                               // 769
    }                                                                                                        // 770
  }                                                                                                          // 771
}                                                                                                            // 772
                                                                                                             // 773
allowDevAccounts = function () {                                                                             // 774
  var setting = Settings.findOne({_id: "devAccounts"});                                                      // 775
  if (setting) {                                                                                             // 776
    return setting.value;                                                                                    // 777
  } else {                                                                                                   // 778
    return Meteor.settings && Meteor.settings.public &&                                                      // 779
           Meteor.settings.public.allowDevAccounts;                                                          // 780
  }                                                                                                          // 781
};                                                                                                           // 782
                                                                                                             // 783
roleAssignmentPattern = {                                                                                    // 784
  none : Match.Optional(null),                                                                               // 785
  allAccess: Match.Optional(null),                                                                           // 786
  roleId: Match.Optional(Match.Integer),                                                                     // 787
  addPermissions: Match.Optional([Boolean]),                                                                 // 788
  removePermissions: Match.Optional([Boolean]),                                                              // 789
};                                                                                                           // 790
                                                                                                             // 791
SandstormDb = function () {                                                                                  // 792
  this.collections = {                                                                                       // 793
    // Direct access to underlying collections. DEPRECATED.                                                  // 794
    //                                                                                                       // 795
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove            // 796
    //   direct access to the collections.                                                                   // 797
                                                                                                             // 798
    packages: Packages,                                                                                      // 799
    devPackages: DevPackages,                                                                                // 800
    userActions: UserActions,                                                                                // 801
    grains: Grains,                                                                                          // 802
    contacts: Contacts,                                                                                      // 803
    sessions: Sessions,                                                                                      // 804
    signupKeys: SignupKeys,                                                                                  // 805
    activityStats: ActivityStats,                                                                            // 806
    deleteStats: DeleteStats,                                                                                // 807
    fileTokens: FileTokens,                                                                                  // 808
    apiTokens: ApiTokens,                                                                                    // 809
    notifications: Notifications,                                                                            // 810
    statsTokens: StatsTokens,                                                                                // 811
    misc: Misc,                                                                                              // 812
    settings: Settings,                                                                                      // 813
    appIndex: AppIndex,                                                                                      // 814
    keybaseProfiles: KeybaseProfiles,                                                                        // 815
                                                                                                             // 816
    // Intentionally omitted:                                                                                // 817
    // - Migrations, since it's used only within this package.                                               // 818
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.          // 819
  };                                                                                                         // 820
};                                                                                                           // 821
                                                                                                             // 822
// TODO(cleanup): These methods should not be defined freestanding and should use collection                 // 823
//   objects created in SandstormDb's constructor rather than globals.                                       // 824
                                                                                                             // 825
_.extend(SandstormDb.prototype, {                                                                            // 826
  isDemoUser: isDemoUser,                                                                                    // 827
  isSignedUp: isSignedUp,                                                                                    // 828
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                        // 829
  isUserOverQuota: isUserOverQuota,                                                                          // 830
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                                    // 831
  isAdmin: isAdmin,                                                                                          // 832
  isAdminById: isAdminById,                                                                                  // 833
  findAdminUserForToken: findAdminUserForToken,                                                              // 834
  matchWildcardHost: matchWildcardHost,                                                                      // 835
  makeWildcardHost: makeWildcardHost,                                                                        // 836
  allowDevAccounts: allowDevAccounts,                                                                        // 837
  roleAssignmentPattern: roleAssignmentPattern,                                                              // 838
});                                                                                                          // 839
                                                                                                             // 840
if (Meteor.isServer) {                                                                                       // 841
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                               // 842
}                                                                                                            // 843
                                                                                                             // 844
// =======================================================================================                   // 845
// Below this point are newly-written or refactored functions.                                               // 846
                                                                                                             // 847
_.extend(SandstormDb.prototype, {                                                                            // 848
  getUser: function getUser (userId) {                                                                       // 849
    check(userId, Match.OneOf(String, undefined, null));                                                     // 850
    if (userId) {                                                                                            // 851
      return Meteor.users.findOne(userId);                                                                   // 852
    }                                                                                                        // 853
  },                                                                                                         // 854
                                                                                                             // 855
  getIdentity: function getIdentity (identityId) {                                                           // 856
    check(identityId, String);                                                                               // 857
    var identity = Meteor.users.findOne({_id: identityId});                                                  // 858
    if (identity) {                                                                                          // 859
      SandstormDb.fillInProfileDefaults(identity);                                                           // 860
      SandstormDb.fillInIntrinsicName(identity);                                                             // 861
      SandstormDb.fillInPictureUrl(identity);                                                                // 862
      return identity;                                                                                       // 863
    }                                                                                                        // 864
  },                                                                                                         // 865
                                                                                                             // 866
  userHasIdentity: function (userId, identityId) {                                                           // 867
    check(userId, String);                                                                                   // 868
    check(identityId, String);                                                                               // 869
                                                                                                             // 870
    if (userId === identityId) return true;                                                                  // 871
                                                                                                             // 872
    var user = Meteor.users.findOne(userId);                                                                 // 873
    return SandstormDb.getUserIdentityIds(user).indexOf(identityId) != -1;                                   // 874
  },                                                                                                         // 875
                                                                                                             // 876
  userGrains: function userGrains (userId) {                                                                 // 877
    check(userId, Match.OneOf(String, undefined, null));                                                     // 878
    return this.collections.grains.find({userId: userId});                                                   // 879
  },                                                                                                         // 880
                                                                                                             // 881
  currentUserGrains: function currentUserGrains () {                                                         // 882
    return this.userGrains(Meteor.userId());                                                                 // 883
  },                                                                                                         // 884
                                                                                                             // 885
  getGrain: function getGrain (grainId) {                                                                    // 886
    check(grainId, String);                                                                                  // 887
    return this.collections.grains.findOne(grainId);                                                         // 888
  },                                                                                                         // 889
                                                                                                             // 890
  userApiTokens: function userApiTokens (userId) {                                                           // 891
    check(userId, Match.OneOf(String, undefined, null));                                                     // 892
    var identityIds = SandstormDb.getUserIdentityIds(this.getUser(userId));                                  // 893
    return this.collections.apiTokens.find({'owner.user.identityId': {$in: identityIds}});                   // 894
  },                                                                                                         // 895
                                                                                                             // 896
  currentUserApiTokens: function currentUserApiTokens () {                                                   // 897
    return this.userApiTokens(Meteor.userId());                                                              // 898
  },                                                                                                         // 899
                                                                                                             // 900
  userActions: function userActions (user) {                                                                 // 901
    return this.collections.userActions.find({userId: user});                                                // 902
  },                                                                                                         // 903
                                                                                                             // 904
  currentUserActions: function currentUserActions () {                                                       // 905
    return this.userActions(Meteor.userId());                                                                // 906
  },                                                                                                         // 907
                                                                                                             // 908
  iconSrcForPackage: function iconSrcForPackage (pkg, usage) {                                               // 909
    return Identicon.iconSrcForPackage(pkg, usage, this.makeWildcardHost("static"));                         // 910
  },                                                                                                         // 911
                                                                                                             // 912
  getPlan: function (id) {                                                                                   // 913
    check(id, String);                                                                                       // 914
    var plan = Plans.findOne(id);                                                                            // 915
    if (!plan) {                                                                                             // 916
      throw new Error("no such plan: " + id);                                                                // 917
    }                                                                                                        // 918
    return plan;                                                                                             // 919
  },                                                                                                         // 920
                                                                                                             // 921
  listPlans: function () {                                                                                   // 922
    return Plans.find({}, {sort: {price: 1}});                                                               // 923
  },                                                                                                         // 924
                                                                                                             // 925
  getMyPlan: function () {                                                                                   // 926
    var user = Meteor.user();                                                                                // 927
    return user && Plans.findOne(user.plan || "free");                                                       // 928
  },                                                                                                         // 929
                                                                                                             // 930
  getMyReferralBonus: function(user) {                                                                       // 931
    // This function is called from the server and from the client, similar to getMyPlan().                  // 932
    //                                                                                                       // 933
    // When called from the server, calculate the user's actual referral bonus. We use this                  // 934
    // elsewhere to store a value in user.pseudoReferralBonus.                                               // 935
    //                                                                                                       // 936
    // When called from the client, return the value of user.pseudoReferralBonus (if it exists).             // 937
    if (Meteor.isClient) {                                                                                   // 938
      // If called on the client side, always use the currently logged-in user.                              // 939
      user = Meteor.user();                                                                                  // 940
                                                                                                             // 941
      if (user && user.pseudoReferralBonus) {                                                                // 942
        return user.pseudoReferralBonus;                                                                     // 943
      }                                                                                                      // 944
      // If we get to this, the subscriptions haven't arrived yet.                                           // 945
      var noBonus = {grains: 0, storage: 0};                                                                 // 946
      return noBonus;                                                                                        // 947
    }                                                                                                        // 948
    if (Meteor.isServer) {                                                                                   // 949
      var x = calculateReferralBonus(user._id, user.plan);                                                   // 950
      return x;                                                                                              // 951
    }                                                                                                        // 952
  },                                                                                                         // 953
                                                                                                             // 954
  getMyUsage: function (user) {                                                                              // 955
    user = user || Meteor.user();                                                                            // 956
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                                     // 957
      if (Meteor.isClient) {                                                                                 // 958
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly              // 959
        // not reactive.                                                                                     // 960
        return user.pseudoUsage;                                                                             // 961
      } else {                                                                                               // 962
        return {                                                                                             // 963
          grains: Grains.find({userId: user._id}).count(),                                                   // 964
          storage: user.storageUsage || 0,                                                                   // 965
          compute: 0   // not tracked yet                                                                    // 966
        };                                                                                                   // 967
      }                                                                                                      // 968
    } else {                                                                                                 // 969
      return {grains: 0, storage: 0, compute: 0};                                                            // 970
    }                                                                                                        // 971
  },                                                                                                         // 972
                                                                                                             // 973
  isUninvitedFreeUser: function () {                                                                         // 974
    if (!Meteor.settings.public.allowUninvited) return false;                                                // 975
                                                                                                             // 976
    var user = Meteor.user();                                                                                // 977
    return user && !user.expires && (!user.plan || user.plan === "free");                                    // 978
  },                                                                                                         // 979
                                                                                                             // 980
  getSetting: function (name) {                                                                              // 981
    var setting = Settings.findOne(name);                                                                    // 982
    return setting && setting.value;                                                                         // 983
  },                                                                                                         // 984
                                                                                                             // 985
  addUserActions: function(packageId) {                                                                      // 986
    //TODO(cleanup): implement this with meteor methods rather than client-side inserts/removes.             // 987
    var pack = Packages.findOne(packageId);                                                                  // 988
    if (pack) {                                                                                              // 989
      // Remove old versions.                                                                                // 990
      UserActions.find({userId: Meteor.userId(), appId: pack.appId})                                         // 991
          .forEach(function (action) {                                                                       // 992
        UserActions.remove(action._id);                                                                      // 993
      });                                                                                                    // 994
                                                                                                             // 995
      // Install new.                                                                                        // 996
      var actions = pack.manifest.actions;                                                                   // 997
      for (i in actions) {                                                                                   // 998
        var action = actions[i];                                                                             // 999
        if ("none" in action.input) {                                                                        // 1000
          var userAction = {                                                                                 // 1001
            userId: Meteor.userId(),                                                                         // 1002
            packageId: pack._id,                                                                             // 1003
            appId: pack.appId,                                                                               // 1004
            appTitle: pack.manifest.appTitle,                                                                // 1005
            appMarketingVersion: pack.manifest.appMarketingVersion,                                          // 1006
            appVersion: pack.manifest.appVersion,                                                            // 1007
            title: action.title,                                                                             // 1008
            nounPhrase: action.nounPhrase,                                                                   // 1009
            command: action.command                                                                          // 1010
          };                                                                                                 // 1011
          UserActions.insert(userAction);                                                                    // 1012
        } else {                                                                                             // 1013
          // TODO(someday):  Implement actions with capability inputs.                                       // 1014
        }                                                                                                    // 1015
      }                                                                                                      // 1016
                                                                                                             // 1017
      Meteor.call("deleteUnusedPackages", pack.appId);                                                       // 1018
    }                                                                                                        // 1019
  },                                                                                                         // 1020
                                                                                                             // 1021
  sendAdminNotification: function (message, link) {                                                          // 1022
    Meteor.users.find({isAdmin: true}, {fields: {_id: 1}}).forEach(function (user) {                         // 1023
      Notifications.insert({                                                                                 // 1024
        admin: {                                                                                             // 1025
          action: link,                                                                                      // 1026
          type: "reportStats",                                                                               // 1027
        },                                                                                                   // 1028
        userId: user._id,                                                                                    // 1029
        text: {defaultText: message},                                                                        // 1030
        timestamp: new Date(),                                                                               // 1031
        isUnread: true,                                                                                      // 1032
      });                                                                                                    // 1033
    });                                                                                                      // 1034
  },                                                                                                         // 1035
                                                                                                             // 1036
  getKeybaseProfile: function (keyFingerprint) {                                                             // 1037
    return this.collections.keybaseProfiles.findOne(keyFingerprint) || {};                                   // 1038
  },                                                                                                         // 1039
});                                                                                                          // 1040
                                                                                                             // 1041
var appNameFromPackage = function(packageObj) {                                                              // 1042
  // This function takes a Package object from Mongo and returns an                                          // 1043
  // app title.                                                                                              // 1044
  var manifest = packageObj.manifest;                                                                        // 1045
  if (!manifest) return packageObj.appId || packageObj._id || "unknown";                                     // 1046
  var action = manifest.actions[0];                                                                          // 1047
  appName = (manifest.appTitle && manifest.appTitle.defaultText) ||                                          // 1048
    appNameFromActionName(action.title.defaultText);                                                         // 1049
  return appName;                                                                                            // 1050
};                                                                                                           // 1051
                                                                                                             // 1052
var appNameFromActionName = function(name) {                                                                 // 1053
  // Hack: Historically we only had action titles, like "New Etherpad Document", not app                     // 1054
  //   titles. But for this UI we want app titles. As a transitionary measure, try to                        // 1055
  //   derive the app title from the action title.                                                           // 1056
  // TODO(cleanup): Get rid of this once apps have real titles.                                              // 1057
  if (!name) {                                                                                               // 1058
    return "(unnamed)";                                                                                      // 1059
  }                                                                                                          // 1060
  if (name.lastIndexOf("New ", 0) === 0) {                                                                   // 1061
    name = name.slice(4);                                                                                    // 1062
  }                                                                                                          // 1063
  if (name.lastIndexOf("Hacker CMS", 0) === 0) {                                                             // 1064
    name = "Hacker CMS";                                                                                     // 1065
  } else {                                                                                                   // 1066
    var space = name.indexOf(" ");                                                                           // 1067
    if (space > 0) {                                                                                         // 1068
      name = name.slice(0, space);                                                                           // 1069
    }                                                                                                        // 1070
  }                                                                                                          // 1071
  return name;                                                                                               // 1072
};                                                                                                           // 1073
                                                                                                             // 1074
var appShortDescriptionFromPackage = function (pkg) {                                                        // 1075
  return pkg && pkg.manifest && pkg.manifest.metadata &&                                                     // 1076
         pkg.manifest.metadata.shortDescription &&                                                           // 1077
         pkg.manifest.metadata.shortDescription.defaultText;                                                 // 1078
};                                                                                                           // 1079
                                                                                                             // 1080
var nounPhraseForActionAndAppTitle = function(action, appTitle) {                                            // 1081
  // A hack to deal with legacy apps not including fields in their manifests.                                // 1082
  // I look forward to the day I can remove most of this code.                                               // 1083
  // Attempt to figure out the appropriate noun that this action will create.                                // 1084
  // Use an explicit noun phrase is one is available.  Apps should add these in the future.                  // 1085
  if (action.nounPhrase) return action.nounPhrase.defaultText;                                               // 1086
  // Otherwise, try to guess one from the structure of the action title field                                // 1087
  if (action.title && action.title.defaultText) {                                                            // 1088
    var text = action.title.defaultText;                                                                     // 1089
    // Strip a leading "New "                                                                                // 1090
    if (text.lastIndexOf("New ", 0) === 0) {                                                                 // 1091
      var candidate = text.slice(4);                                                                         // 1092
      // Strip a leading appname too, if provided                                                            // 1093
      if (candidate.lastIndexOf(appTitle, 0) === 0) {                                                        // 1094
        var newCandidate = candidate.slice(appTitle.length);                                                 // 1095
        // Unless that leaves you with no noun, in which case, use "grain"                                   // 1096
        if (newCandidate.length > 0) {                                                                       // 1097
          return newCandidate.toLowerCase();                                                                 // 1098
        } else {                                                                                             // 1099
          return "grain";                                                                                    // 1100
        }                                                                                                    // 1101
      }                                                                                                      // 1102
      return candidate.toLowerCase();                                                                        // 1103
    }                                                                                                        // 1104
    // Some other verb phrase was given.  Just use it verbatim, and hope the app author updates              // 1105
    // the package soon.                                                                                     // 1106
    return text;                                                                                             // 1107
  } else {                                                                                                   // 1108
    return "grain";                                                                                          // 1109
  }                                                                                                          // 1110
};                                                                                                           // 1111
                                                                                                             // 1112
// Static methods on SandstormDb that don't need an instance.                                                // 1113
// Largely things that deal with backwards-compatibility.                                                    // 1114
_.extend(SandstormDb, {                                                                                      // 1115
  appNameFromActionName: appNameFromActionName,                                                              // 1116
  appNameFromPackage: appNameFromPackage,                                                                    // 1117
  appShortDescriptionFromPackage: appShortDescriptionFromPackage,                                            // 1118
  nounPhraseForActionAndAppTitle: nounPhraseForActionAndAppTitle,                                            // 1119
});                                                                                                          // 1120
                                                                                                             // 1121
if (Meteor.isServer) {                                                                                       // 1122
  var Crypto = Npm.require("crypto");                                                                        // 1123
  var ContentType = Npm.require("content-type");                                                             // 1124
  var Zlib = Npm.require("zlib");                                                                            // 1125
                                                                                                             // 1126
  var replicaNumber = Meteor.settings.replicaNumber || 0;                                                    // 1127
                                                                                                             // 1128
  var computeStagger = function (n) {                                                                        // 1129
    // Compute a fraction in the range [0, 1) such that, for any natural number k, the values                // 1130
    // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.                 // 1131
    // The sequence looks like:                                                                              // 1132
    //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                                     // 1133
    //                                                                                                       // 1134
    // We use this to determine how we'll stagger periodic events performed by this replica.                 // 1135
    // Notice that this allows us to compute a stagger which is independent of the number of                 // 1136
    // front-end replicas present; we can add more replicas to the end without affecting how the             // 1137
    // earlier ones schedule their events.                                                                   // 1138
    var denom = 1;                                                                                           // 1139
    while (denom <= n) denom <<= 1;                                                                          // 1140
    var num = n * 2 - denom + 1;                                                                             // 1141
    return num / denom;                                                                                      // 1142
  }                                                                                                          // 1143
                                                                                                             // 1144
  var stagger = computeStagger(replicaNumber);                                                               // 1145
                                                                                                             // 1146
  SandstormDb.periodicCleanup = function (intervalMs, callback) {                                            // 1147
    // Register a database cleanup function than should run periodically, roughly once every                 // 1148
    // interval of the given length.                                                                         // 1149
    //                                                                                                       // 1150
    // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be              // 1151
    // scaled appropriately on the assumption that more data is being generated demanding more               // 1152
    // frequent cleanups.                                                                                    // 1153
                                                                                                             // 1154
    check(intervalMs, Number);                                                                               // 1155
    check(callback, Function);                                                                               // 1156
                                                                                                             // 1157
    if (intervalMs < 120000) {                                                                               // 1158
      throw new Error("less than 2-minute cleanup interval seems too fast; " +                               // 1159
                      "are you using the right units?");                                                     // 1160
    }                                                                                                        // 1161
                                                                                                             // 1162
    // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that              // 1163
    // the schedule is independent of the exact startup time.                                                // 1164
    var first = intervalMs - Date.now() % intervalMs;                                                        // 1165
                                                                                                             // 1166
    // Stagger cleanups across replicas so that we don't have all replicas trying to clean the               // 1167
    // same data at the same time.                                                                           // 1168
    first += Math.floor(intervalMs * computeStagger(replicaNumber));                                         // 1169
                                                                                                             // 1170
    // If the stagger put us more than an interval away from now, back up.                                   // 1171
    if (first > intervalMs) first -= intervalMs;                                                             // 1172
                                                                                                             // 1173
    Meteor.setTimeout(function () {                                                                          // 1174
      callback();                                                                                            // 1175
      Meteor.setInterval(callback, intervalMs);                                                              // 1176
    }, first);                                                                                               // 1177
  }                                                                                                          // 1178
                                                                                                             // 1179
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.                  // 1180
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                          // 1181
                                                                                                             // 1182
  var BufferSmallerThan = function (limit) {                                                                 // 1183
    return Match.Where(function (buf) {                                                                      // 1184
      check(buf, Buffer);                                                                                    // 1185
      return buf.length < limit;                                                                             // 1186
    });                                                                                                      // 1187
  }                                                                                                          // 1188
                                                                                                             // 1189
  var DatabaseId = Match.Where(function (s) {                                                                // 1190
    check(s, String);                                                                                        // 1191
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                                     // 1192
  });                                                                                                        // 1193
                                                                                                             // 1194
  addStaticAsset = function (metadata, content) {                                                            // 1195
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it             // 1196
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.          // 1197
                                                                                                             // 1198
    if (typeof content === "string" && !metadata.encoding) {                                                 // 1199
      content = gzipSync(new Buffer(content, "utf8"));                                                       // 1200
      metadata.encoding = "gzip";                                                                            // 1201
    }                                                                                                        // 1202
                                                                                                             // 1203
    check(metadata, {                                                                                        // 1204
      mimeType: String,                                                                                      // 1205
      encoding: Match.Optional("gzip")                                                                       // 1206
    });                                                                                                      // 1207
    check(content, BufferSmallerThan(1 << 20));                                                              // 1208
                                                                                                             // 1209
    // Validate content type.                                                                                // 1210
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                            // 1211
                                                                                                             // 1212
    var hasher = Crypto.createHash("sha256");                                                                // 1213
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                              // 1214
    hasher.update(content);                                                                                  // 1215
    var hash = hasher.digest("base64");                                                                      // 1216
                                                                                                             // 1217
    var existing = StaticAssets.findAndModify({                                                              // 1218
      query: {hash: hash, refcount: {$gte: 1}},                                                              // 1219
      update: {$inc: {refcount: 1}},                                                                         // 1220
      fields: {_id: 1, refcount: 1},                                                                         // 1221
    });                                                                                                      // 1222
    if (existing) {                                                                                          // 1223
      return existing._id;                                                                                   // 1224
    }                                                                                                        // 1225
                                                                                                             // 1226
    return StaticAssets.insert(_.extend({                                                                    // 1227
      hash: hash,                                                                                            // 1228
      content: content,                                                                                      // 1229
      refcount: 1                                                                                            // 1230
    }, metadata));                                                                                           // 1231
  }                                                                                                          // 1232
                                                                                                             // 1233
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                                     // 1234
                                                                                                             // 1235
  SandstormDb.prototype.refStaticAsset = function (id) {                                                     // 1236
    // Increment the refcount on an existing static asset.                                                   // 1237
    //                                                                                                       // 1238
    // You must call this BEFORE adding the new reference to the DB, in case of failure between              // 1239
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1240
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1241
                                                                                                             // 1242
    check(id, String);                                                                                       // 1243
                                                                                                             // 1244
    var existing = StaticAssets.findAndModify({                                                              // 1245
      query: {hash: hash},                                                                                   // 1246
      update: {$inc: {refcount: 1}},                                                                         // 1247
      fields: {_id: 1, refcount: 1},                                                                         // 1248
    });                                                                                                      // 1249
    if (!existing) {                                                                                         // 1250
      throw new Error("refStaticAsset() called on asset that doesn't exist");                                // 1251
    }                                                                                                        // 1252
  }                                                                                                          // 1253
                                                                                                             // 1254
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                                   // 1255
    // Decrement refcount on a static asset and delete if it has reached zero.                               // 1256
    //                                                                                                       // 1257
    // You must call this AFTER removing the reference from the DB, in case of failure between               // 1258
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1259
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1260
                                                                                                             // 1261
    check(id, String);                                                                                       // 1262
                                                                                                             // 1263
    var existing = StaticAssets.findAndModify({                                                              // 1264
      query: {_id: id},                                                                                      // 1265
      update: {$inc: {refcount: -1}},                                                                        // 1266
      fields: {_id: 1, refcount: 1},                                                                         // 1267
      new: true,                                                                                             // 1268
    });                                                                                                      // 1269
    if (!existing) {                                                                                         // 1270
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);               // 1271
    } else if (existing.refcount <= 0) {                                                                     // 1272
      StaticAssets.remove({_id: existing._id});                                                              // 1273
    }                                                                                                        // 1274
  }                                                                                                          // 1275
                                                                                                             // 1276
  SandstormDb.prototype.getStaticAsset = function (id) {                                                     // 1277
    // Get a static asset's mimeType, encoding, and raw content.                                             // 1278
                                                                                                             // 1279
    check(id, String);                                                                                       // 1280
                                                                                                             // 1281
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});          // 1282
    if (asset) {                                                                                             // 1283
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy              // 1284
      //   here.                                                                                             // 1285
      asset.content = new Buffer(asset.content);                                                             // 1286
    }                                                                                                        // 1287
    return asset;                                                                                            // 1288
  }                                                                                                          // 1289
                                                                                                             // 1290
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                                // 1291
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: DatabaseId}});                          // 1292
                                                                                                             // 1293
    return AssetUploadTokens.insert({                                                                        // 1294
      purpose: purpose,                                                                                      // 1295
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                               // 1296
    });                                                                                                      // 1297
  }                                                                                                          // 1298
                                                                                                             // 1299
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                                 // 1300
    // Indicates that the given asset upload has completed. It will be removed and its purpose               // 1301
    // returned. If no matching upload exists, returns undefined.                                            // 1302
                                                                                                             // 1303
    check(id, String);                                                                                       // 1304
                                                                                                             // 1305
    var upload = AssetUploadTokens.findAndModify({                                                           // 1306
      query: {_id: id},                                                                                      // 1307
      remove: true                                                                                           // 1308
    });                                                                                                      // 1309
                                                                                                             // 1310
    if (upload.expires.valueOf() < Date.now()) {                                                             // 1311
      return undefined;  // already expired                                                                  // 1312
    } else {                                                                                                 // 1313
      return upload.purpose;                                                                                 // 1314
    }                                                                                                        // 1315
  }                                                                                                          // 1316
                                                                                                             // 1317
  function cleanupExpiredAssetUploads() {                                                                    // 1318
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                                  // 1319
  }                                                                                                          // 1320
                                                                                                             // 1321
  // Cleanup tokens every hour.                                                                              // 1322
  SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                          // 1323
                                                                                                             // 1324
  var packageCache = {};                                                                                     // 1325
  // Package info is immutable. Let's cache to save on mongo queries.                                        // 1326
                                                                                                             // 1327
  SandstormDb.prototype.getPackage = function (packageId) {                                                  // 1328
    // Get the given package record. Since package info is immutable, cache the data in the server           // 1329
    // to reduce mongo query overhead, since it turns out we have to fetch specific packages a               // 1330
    // lot.                                                                                                  // 1331
                                                                                                             // 1332
    if (packageId in packageCache) {                                                                         // 1333
      return packageCache[packageId];                                                                        // 1334
    }                                                                                                        // 1335
                                                                                                             // 1336
    var pkg = Packages.findOne(packageId);                                                                   // 1337
    if (pkg && pkg.status === "ready") {                                                                     // 1338
      packageCache[packageId] = pkg;                                                                         // 1339
    }                                                                                                        // 1340
    return pkg;                                                                                              // 1341
  }                                                                                                          // 1342
                                                                                                             // 1343
  SandstormDb.prototype.sendAppUpdateNotifications = function (appId, packageId, name,                       // 1344
                                                               versionNumber, marketingVersion) {            // 1345
    var db = this;                                                                                           // 1346
    var actions = db.collections.userActions.find({appId: appId, appVersion: {$lt: versionNumber}},          // 1347
      {fields: {userId: 1}});                                                                                // 1348
    actions.forEach(function (action) {                                                                      // 1349
      var userId = action.userId;                                                                            // 1350
      var updater = {                                                                                        // 1351
        userId: userId,                                                                                      // 1352
        timestamp: new Date(),                                                                               // 1353
        isUnread: true,                                                                                      // 1354
      };                                                                                                     // 1355
                                                                                                             // 1356
      // Set only the appId that we care about. Use mongo's dot notation to specify only a single            // 1357
      // field inside of an object to update                                                                 // 1358
      updater["appUpdates." + appId] = {                                                                     // 1359
        marketingVersion: marketingVersion,                                                                  // 1360
        packageId: packageId,                                                                                // 1361
        name: name,                                                                                          // 1362
        version: versionNumber,                                                                              // 1363
      };                                                                                                     // 1364
      db.collections.notifications.upsert({userId: userId}, { $set: updater});                               // 1365
    });                                                                                                      // 1366
                                                                                                             // 1367
    db.collections.appIndex.update({_id: appId}, {$set: {hasSentNotifications: true}});                      // 1368
                                                                                                             // 1369
    // In the case where we replaced a previous notification and that was the only reference to the          // 1370
    // package, we need to clean it up                                                                       // 1371
    Meteor.call("deleteUnusedPackages", appId);                                                              // 1372
  };                                                                                                         // 1373
                                                                                                             // 1374
  SandstormDb.prototype.upgradeGrains =  function (appId, version, packageId, backend) {                     // 1375
    check(appId, String);                                                                                    // 1376
    check(version, Match.Integer);                                                                           // 1377
    check(packageId, String);                                                                                // 1378
                                                                                                             // 1379
    var selector = {                                                                                         // 1380
      userId: Meteor.userId(),                                                                               // 1381
      appId: appId,                                                                                          // 1382
      appVersion: { $lte: version },                                                                         // 1383
      packageId: { $ne: packageId }                                                                          // 1384
    };                                                                                                       // 1385
                                                                                                             // 1386
    if (!this.isSimulation) {                                                                                // 1387
      Grains.find(selector).forEach(function (grain) {                                                       // 1388
        backend.shutdownGrain(grain._id, grain.userId);                                                      // 1389
      });                                                                                                    // 1390
    }                                                                                                        // 1391
                                                                                                             // 1392
    Grains.update(selector, { $set: { appVersion: version, packageId: packageId }}, {multi: true});          // 1393
  };                                                                                                         // 1394
                                                                                                             // 1395
  SandstormDb.prototype.startInstall = function (packageId, url, retryFailed, isAutoUpdated) {               // 1396
    // Mark package for possible installation.                                                               // 1397
                                                                                                             // 1398
    var fields = {status: "download", progress: 0, url: url, isAutoUpdated: !!isAutoUpdated};                // 1399
                                                                                                             // 1400
    if (retryFailed) {                                                                                       // 1401
      Packages.update({_id: packageId, status: "failed"}, {$set: fields});                                   // 1402
    } else {                                                                                                 // 1403
      try {                                                                                                  // 1404
        fields._id = packageId;                                                                              // 1405
        Packages.insert(fields);                                                                             // 1406
      } catch (err) {                                                                                        // 1407
        console.error("Simultaneous startInstall()s?", err.stack);                                           // 1408
      }                                                                                                      // 1409
    }                                                                                                        // 1410
  };                                                                                                         // 1411
                                                                                                             // 1412
  var ValidKeyFingerprint = Match.Where(function (keyFingerprint) {                                          // 1413
    check(keyFingerprint, String);                                                                           // 1414
    return !!keyFingerprint.match(/^[0-9A-F]{40}$/);                                                         // 1415
  });                                                                                                        // 1416
                                                                                                             // 1417
  SandstormDb.prototype.updateKeybaseProfileAsync = function (keyFingerprint) {                              // 1418
    // Asynchronously fetch the given Keybase profile and populate the KeybaseProfiles collection.           // 1419
                                                                                                             // 1420
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1421
                                                                                                             // 1422
    console.log("fetching keybase", keyFingerprint);                                                         // 1423
                                                                                                             // 1424
    HTTP.get(                                                                                                // 1425
        "https://keybase.io/_/api/1.0/user/lookup.json?key_fingerprint=" + keyFingerprint +                  // 1426
        "&fields=basics,profile,proofs_summary", {                                                           // 1427
      timeout: 5000                                                                                          // 1428
    }, function (err, keybaseResponse) {                                                                     // 1429
      if (err) {                                                                                             // 1430
        console.log("keybase lookup error:", err.stack);                                                     // 1431
        return;                                                                                              // 1432
      }                                                                                                      // 1433
                                                                                                             // 1434
      if (!keybaseResponse.data) {                                                                           // 1435
        console.log("keybase didn't return JSON? Headers:", keybaseResponse.headers);                        // 1436
        return;                                                                                              // 1437
      }                                                                                                      // 1438
                                                                                                             // 1439
      var profile = (keybaseResponse.data.them || [])[0];                                                    // 1440
                                                                                                             // 1441
      if (profile) {                                                                                         // 1442
        var record = {                                                                                       // 1443
          displayName: (profile.profile || {}).full_name,                                                    // 1444
          handle: (profile.basics || {}).username,                                                           // 1445
          proofs: (profile.proofs_summary || {}).all || []                                                   // 1446
        };                                                                                                   // 1447
                                                                                                             // 1448
        record.proofs.forEach(function (proof) {                                                             // 1449
          // Remove potentially Mongo-incompatible stuff. (Currently Keybase returns nothing that            // 1450
          // this would filter.)                                                                             // 1451
          for (var field in proof) {                                                                         // 1452
            // Don't allow field names containing '.' or '$'. Also don't allow sub-objects mainly            // 1453
            // because I'm too lazy to check the field names recursively (and Keybase doesn't                // 1454
            // return any objects anyway).                                                                   // 1455
            if (field.match(/[.$]/) || typeof(proof[field]) === "object") {                                  // 1456
              delete proof[field];                                                                           // 1457
            }                                                                                                // 1458
          }                                                                                                  // 1459
                                                                                                             // 1460
          // Indicate not verified.                                                                          // 1461
          // TODO(security): Asynchronously verify proofs. Presumably we can borrow code from the            // 1462
          //   Keybase node-based CLI.                                                                       // 1463
          proof.status = "unverified";                                                                       // 1464
        });                                                                                                  // 1465
                                                                                                             // 1466
        KeybaseProfiles.update(keyFingerprint, {$set: record}, {upsert: true});                              // 1467
      } else {                                                                                               // 1468
        // Keybase reports no match, so remove what we know of this user. We don't want to remove            // 1469
        // the item entirely from the cache as this will cause us to repeatedly re-fetch the data            // 1470
        // from Keybase.                                                                                     // 1471
        //                                                                                                   // 1472
        // TODO(someday): We could perhaps keep the proofs if we can still verify them directly,             // 1473
        //   but at present we don't have the ability to verify proofs.                                      // 1474
        KeybaseProfiles.update(keyFingerprint,                                                               // 1475
            {$unset: {displayName: "", handle: "", proofs: ""}}, {upsert: true});                            // 1476
      }                                                                                                      // 1477
    });                                                                                                      // 1478
  };                                                                                                         // 1479
                                                                                                             // 1480
  SandstormDb.prototype.deleteUnusedAccount = function(backend, identityId) {                                // 1481
    // If there is an *unused* account that has `identityId` as a login identity, deletes it.                // 1482
                                                                                                             // 1483
    check(identityId, String);                                                                               // 1484
    var account = Meteor.users.findOne({"loginIdentities.id": identityId});                                  // 1485
    if (account &&                                                                                           // 1486
        account.loginIdentities.length == 1 &&                                                               // 1487
        account.nonloginIdentities.length == 0 &&                                                            // 1488
        !Grains.findOne({userId: account._id}) &&                                                            // 1489
        !ApiTokens.findOne({accountId: account._id}) &&                                                      // 1490
        (!account.plan || account.plan === "free") &&                                                        // 1491
        !account.payments &&                                                                                 // 1492
        !Notifications.findOne({userId: account._id}) &&                                                     // 1493
        !Contacts.findOne({ownerId: account._id})) {                                                         // 1494
      Meteor.users.remove({_id: account._id});                                                               // 1495
      backend.deleteUser(account._id);                                                                       // 1496
    }                                                                                                        // 1497
  }                                                                                                          // 1498
                                                                                                             // 1499
  Meteor.publish("keybaseProfile", function (keyFingerprint) {                                               // 1500
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1501
    var db = this.connection.sandstormDb;                                                                    // 1502
                                                                                                             // 1503
    var cursor = db.collections.keybaseProfiles.find(keyFingerprint);                                        // 1504
    if (cursor.count() === 0) {                                                                              // 1505
      // Fire off async update.                                                                              // 1506
      db.updateKeybaseProfileAsync(keyFingerprint);                                                          // 1507
    }                                                                                                        // 1508
                                                                                                             // 1509
    return cursor;                                                                                           // 1510
  });                                                                                                        // 1511
                                                                                                             // 1512
  Meteor.publish("appIndex", function (appId) {                                                              // 1513
    check(appId, String);                                                                                    // 1514
    var db = this.connection.sandstormDb;                                                                    // 1515
    var cursor = db.collections.appIndex.find({_id: appId});                                                 // 1516
    return cursor;                                                                                           // 1517
  });                                                                                                        // 1518
                                                                                                             // 1519
  Meteor.publish("userPackages", function() {                                                                // 1520
    // Users should be able to see packages that are either:                                                 // 1521
    // 1. referenced by one of their userActions                                                             // 1522
    // 2. referenced by one of their grains                                                                  // 1523
    var self = this;                                                                                         // 1524
    var db = this.connection.sandstormDb;                                                                    // 1525
                                                                                                             // 1526
    // Note that package information, once it is in the database, is static. There's no need to              // 1527
    // reactively subscribe to changes to a package since they don't change. It's also unecessary            // 1528
    // to reactively remove a package from the client side when it is removed on the server, or              // 1529
    // when the client stops using it, because the worst case is the client has a small amount               // 1530
    // of extra info on a no-longer-used package held in memory until they refresh Sandstorm.                // 1531
    // So, we implement this as a cache: the first time each package ID shows up among the user's            // 1532
    // stuff, we push the package info to the client, and then we never update it.                           // 1533
    //                                                                                                       // 1534
    // Alternatively, we could subscribe to each individual package query, but this would waste              // 1535
    // lots of server-side resources watching for events that will never happen or don't matter.             // 1536
    var hasPackage = {};                                                                                     // 1537
    var refPackage = function (packageId) {                                                                  // 1538
      // Ignore dev apps.                                                                                    // 1539
      if (packageId.lastIndexOf("dev-", 0) === 0) return;                                                    // 1540
                                                                                                             // 1541
      if (!hasPackage[packageId]) {                                                                          // 1542
        hasPackage[packageId] = true;                                                                        // 1543
        var pkg = db.getPackage(packageId);                                                                  // 1544
        if (pkg) {                                                                                           // 1545
          self.added("packages", packageId, pkg);                                                            // 1546
        } else {                                                                                             // 1547
          console.error(                                                                                     // 1548
              "shouldn't happen: missing package referenced by user's stuff:", packageId);                   // 1549
        }                                                                                                    // 1550
      }                                                                                                      // 1551
    };                                                                                                       // 1552
                                                                                                             // 1553
    // package source 1: packages referred to by actions                                                     // 1554
    var actions = db.userActions(this.userId);                                                               // 1555
    var actionsHandle = actions.observe({                                                                    // 1556
      added: function(newAction) {                                                                           // 1557
        refPackage(newAction.packageId);                                                                     // 1558
      },                                                                                                     // 1559
      changed: function(oldAction, newAction) {                                                              // 1560
        refPackage(newAction.packageId);                                                                     // 1561
      }                                                                                                      // 1562
    });                                                                                                      // 1563
                                                                                                             // 1564
    // package source 2: packages referred to by grains directly                                             // 1565
    var grains = db.userGrains(this.userId);                                                                 // 1566
    var grainsHandle = grains.observe({                                                                      // 1567
      added: function(newGrain) {                                                                            // 1568
        refPackage(newGrain.packageId);                                                                      // 1569
      },                                                                                                     // 1570
      changed: function(oldGrain, newGrain) {                                                                // 1571
        refPackage(newGrain.packageId);                                                                      // 1572
      }                                                                                                      // 1573
    });                                                                                                      // 1574
                                                                                                             // 1575
    this.onStop(function () {                                                                                // 1576
      actionsHandle.stop();                                                                                  // 1577
      grainsHandle.stop();                                                                                   // 1578
    });                                                                                                      // 1579
    this.ready();                                                                                            // 1580
  });                                                                                                        // 1581
}                                                                                                            // 1582
                                                                                                             // 1583
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/profile.js                                                                          //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var makeIdenticon;                                                                                           // 17
var httpProtocol;                                                                                            // 18
                                                                                                             // 19
if (Meteor.isServer) {                                                                                       // 20
  SandstormDb.ensureSubscriberHasIdentity = function(publishHandler, identityId) {                           // 21
    // Helper for publish functions that need to restrict access based on whether the subscriber             // 22
    // has a given identity linked. Automatically stops the subscription if the user loses the               // 23
    // identity. Returns a boolean indicating whether the user initially has the identity.                   // 24
                                                                                                             // 25
    var userId = publishHandler.userId;                                                                      // 26
    if (userId === identityId) {                                                                             // 27
      return true;                                                                                           // 28
    } else {                                                                                                 // 29
      var hasIdentityCursor =                                                                                // 30
          Meteor.users.find({$or: [{_id: userId, "loginIdentities.id": identityId},                          // 31
                                   {_id: userId, "nonloginIdentities.id": identityId}]});                    // 32
      if (hasIdentityCursor.count() == 0) {                                                                  // 33
        publishHandler.stop();                                                                               // 34
        return false;                                                                                        // 35
      }                                                                                                      // 36
      var handle = hasIdentityCursor.observe({removed: function () { publishHandler.stop(); }});             // 37
      publishHandler.onStop(function () { handle.stop(); });                                                 // 38
      return true;                                                                                           // 39
    }                                                                                                        // 40
  }                                                                                                          // 41
                                                                                                             // 42
  Meteor.publish("identityProfile", function (identityId) {                                                  // 43
    check(identityId, String);                                                                               // 44
    if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return;                                  // 45
                                                                                                             // 46
    return Meteor.users.find({_id: identityId},                                                              // 47
      {fields: {                                                                                             // 48
        "profile":1,                                                                                         // 49
        "unverifiedEmail":1,                                                                                 // 50
        "expires": 1,                                                                                        // 51
                                                                                                             // 52
        "services.dev.name":1,                                                                               // 53
                                                                                                             // 54
        "services.google.id":1,                                                                              // 55
        "services.google.email":1,                                                                           // 56
        "services.google.verified_email":1,                                                                  // 57
        "services.google.name":1,                                                                            // 58
        "services.google.picture":1,                                                                         // 59
        "services.google.gender":1,                                                                          // 60
                                                                                                             // 61
        "services.github.id":1,                                                                              // 62
        "services.github.email":1,                                                                           // 63
        "services.github.emails":1,                                                                          // 64
        "services.github.username":1,                                                                        // 65
                                                                                                             // 66
        "services.email.email":1,                                                                            // 67
      }});                                                                                                   // 68
  }),                                                                                                        // 69
                                                                                                             // 70
  Meteor.publish("accountIdentities", function () {                                                          // 71
    // Maybe this should be folded into the "credentials" subscription?                                      // 72
                                                                                                             // 73
    if (!this.userId) return [];                                                                             // 74
                                                                                                             // 75
    return Meteor.users.find(                                                                                // 76
      {_id: this.userId},                                                                                    // 77
      {fields: {                                                                                             // 78
        "profile": 1,                                                                                        // 79
        "verifiedEmail": 1,                                                                                  // 80
        "loginIdentities": 1,                                                                                // 81
        "nonloginIdentities": 1,                                                                             // 82
        "expires": 1,                                                                                        // 83
        "primaryEmail": 1,                                                                                   // 84
      }});                                                                                                   // 85
  });                                                                                                        // 86
                                                                                                             // 87
  makeIdenticon = function (id) {                                                                            // 88
    // We only make identicons client-side.                                                                  // 89
    return undefined;                                                                                        // 90
  }                                                                                                          // 91
                                                                                                             // 92
  var Url = Npm.require("url");                                                                              // 93
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                                   // 94
} else {                                                                                                     // 95
  var identiconCache = {};                                                                                   // 96
                                                                                                             // 97
  makeIdenticon = function (id) {                                                                            // 98
    // Given a cryptographic hash as input, generate an identicon. We always use the user's                  // 99
    // identity ID as the input hash (the same thing passed to apps as the user ID).                         // 100
                                                                                                             // 101
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so               // 102
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out           // 103
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it              // 104
    // anyway to be safe.                                                                                    // 105
    id = id.slice(0, 32);                                                                                    // 106
                                                                                                             // 107
    if (id in identiconCache) {                                                                              // 108
      return identiconCache[id];                                                                             // 109
    }                                                                                                        // 110
                                                                                                             // 111
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.             // 112
    var data = new Identicon(id, 512).toString();                                                            // 113
    var result = "data:image/png;base64," + data;                                                            // 114
    identiconCache[id] = result;                                                                             // 115
    return result;                                                                                           // 116
  }                                                                                                          // 117
                                                                                                             // 118
  httpProtocol = window.location.protocol;                                                                   // 119
}                                                                                                            // 120
                                                                                                             // 121
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                          // 122
                                                                                                             // 123
function staticAssetUrl(id, staticHost) {                                                                    // 124
  if (id) {                                                                                                  // 125
    return staticHost + "/" + id;                                                                            // 126
  } else {                                                                                                   // 127
    return undefined;                                                                                        // 128
  }                                                                                                          // 129
}                                                                                                            // 130
                                                                                                             // 131
function filterHandle(handle) {                                                                              // 132
  // Convert disallowed letters into underscores, but no more than one underscore in a row.                  // 133
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                                     // 134
}                                                                                                            // 135
                                                                                                             // 136
function emailToHandle(email) {                                                                              // 137
  // Turn an email address into a handle.                                                                    // 138
                                                                                                             // 139
  if (!email) return undefined;                                                                              // 140
                                                                                                             // 141
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters               // 142
  // on the same account.                                                                                    // 143
  var parts = email.split("@");                                                                              // 144
  var base = filterHandle(parts[0].split("+")[0]);                                                           // 145
                                                                                                             // 146
  var domain = (parts[1]||"").split(".");                                                                    // 147
  if (domain[domain.length - 1] === "name") {                                                                // 148
    // Oh, a .name domain. Let's use                                                                         // 149
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                       // 150
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",                  // 151
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],                 // 152
                         base)) {                                                                            // 153
    // This is probably an address at a vanity domain. Use the domain itself as the handle.                  // 154
    base = filterHandle(domain[0]);                                                                          // 155
  }                                                                                                          // 156
                                                                                                             // 157
  return filterHandle(base);                                                                                 // 158
}                                                                                                            // 159
                                                                                                             // 160
SandstormDb.fillInProfileDefaults = function(user) {                                                         // 161
  var profile = user.profile;                                                                                // 162
  if (profile.service === "github") {                                                                        // 163
    profile.name = profile.name || user.services.github.username || "Name Unknown";                          // 164
    profile.handle = profile.handle || filterHandle(user.services.github.username) ||                        // 165
        filterHandle(profile.name);                                                                          // 166
  } else if (profile.service === "google") {                                                                 // 167
    profile.name = profile.name || user.services.google.name || "Name Unknown";                              // 168
    profile.handle = profile.handle || emailToHandle(user.services.google.email) ||                          // 169
        filterHandle(profile.name);                                                                          // 170
    profile.pronoun = profile.pronoun || GENDERS[user.services.google.gender] || "neutral";                  // 171
  } else if (profile.service === "email") {                                                                  // 172
    var email = user.services.email.email                                                                    // 173
    profile.name = profile.name || email.split("@")[0];                                                      // 174
    profile.handle = profile.handle || emailToHandle(email);                                                 // 175
  } else if (profile.service === "dev") {                                                                    // 176
    var lowerCaseName = user.services.dev.name.split(" ")[0].toLowerCase();                                  // 177
    profile.name = profile.name || user.services.dev.name;                                                   // 178
    profile.handle = profile.handle || filterHandle(lowerCaseName);                                          // 179
    profile.pronoun = profile.pronoun ||                                                                     // 180
        (_.contains(["alice", "carol", "eve"], lowerCaseName) ? "female" :                                   // 181
         _.contains(["bob", "dave"], lowerCaseName) ? "male" : "neutral");                                   // 182
  } else if (profile.service === "demo") {                                                                   // 183
    profile.name = profile.name || "Demo User";                                                              // 184
    profile.handle = profile.handle || "demo";                                                               // 185
  } else {                                                                                                   // 186
    throw new Error("unrecognized identity service: ", profile.service);                                     // 187
  }                                                                                                          // 188
                                                                                                             // 189
  profile.pronoun = profile.pronoun || "neutral";                                                            // 190
}                                                                                                            // 191
                                                                                                             // 192
SandstormDb.fillInIntrinsicName = function(user) {                                                           // 193
  var profile = user.profile;                                                                                // 194
  if (profile.service === "github") {                                                                        // 195
    profile.intrinsicName = user.services.github.username;                                                   // 196
  } else if (profile.service === "google") {                                                                 // 197
    profile.intrinsicName = user.services.google.name;                                                       // 198
    user.privateIntrinsicName = user.services.google.email;                                                  // 199
  } else if (profile.service === "email") {                                                                  // 200
    profile.intrinsicName = user.services.email.email                                                        // 201
  } else if (profile.service === "dev") {                                                                    // 202
    profile.intrinsicName = user.services.dev.name;                                                          // 203
  } else if (profile.service === "demo") {                                                                   // 204
    // No intrinsic name.                                                                                    // 205
  } else {                                                                                                   // 206
    throw new Error("unrecognized identity service: ", profile.service);                                     // 207
  }                                                                                                          // 208
}                                                                                                            // 209
                                                                                                             // 210
SandstormDb.getVerifiedEmails = function(identity) {                                                         // 211
  if (identity.services.google && identity.services.google.email &&                                          // 212
      identity.services.google.verified_email) {                                                             // 213
    return [identity.services.google.email];                                                                 // 214
  } else if (identity.services.email) {                                                                      // 215
    return [identity.services.email.email];                                                                  // 216
  } else if (identity.services.github && identity.services.github.emails) {                                  // 217
    return _.pluck(_.filter(identity.services.github.emails,                                                 // 218
                            function(email) { return email.verified; }),                                     // 219
                   "email");                                                                                 // 220
  }                                                                                                          // 221
  return [];                                                                                                 // 222
}                                                                                                            // 223
                                                                                                             // 224
SandstormDb.fillInPictureUrl = function(user) {                                                              // 225
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                         // 226
  user.profile.pictureUrl = staticAssetUrl(user.profile.picture, staticHost) ||                              // 227
    makeIdenticon(user._id);                                                                                 // 228
}                                                                                                            // 229
                                                                                                             // 230
SandstormDb.getUserIdentityIds = function (user) {                                                           // 231
  // Given an account user object, returns an array containing the ID of each identity linked to the         // 232
  // account. Always returns the most recently added login identity first.                                   // 233
  if (user && user.loginIdentities) {                                                                        // 234
    return _.pluck(user.nonloginIdentities.concat(user.loginIdentities), "id").reverse();                    // 235
  } else {                                                                                                   // 236
    return [];                                                                                               // 237
  }                                                                                                          // 238
};                                                                                                           // 239
                                                                                                             // 240
SandstormDb.getUserEmails = function (user) {                                                                // 241
  // Given a user object, returns an array containing all email addresses associated with that user.         // 242
  // Each entry in the array is an object of the form:                                                       // 243
  //     `{email: String, verified: Bool, primary: Optional(Bool)}`                                          // 244
  //                                                                                                         // 245
  // At most one entry in the result has `primary = true`.                                                   // 246
  //                                                                                                         // 247
  // TODO(cleanup): This actually does need to query the database to fetch profile information               // 248
  //   for linked identities, so it probably makes more sense for it to be a non-static method               // 249
  //   on SandstormDb.                                                                                       // 250
                                                                                                             // 251
  var identityIds = SandstormDb.getUserIdentityIds(user);                                                    // 252
  verifiedEmails = {};                                                                                       // 253
  unverifiedEmails = {};                                                                                     // 254
                                                                                                             // 255
  identityIds.forEach(function (id) {                                                                        // 256
    var identity = Meteor.users.findOne({_id: id});                                                          // 257
    if (identity && identity.services) {                                                                     // 258
      SandstormDb.getVerifiedEmails(identity).forEach(function(verifiedEmail) {                              // 259
        if (verifiedEmail) {                                                                                 // 260
          verifiedEmails[verifiedEmail] = true;                                                              // 261
        }                                                                                                    // 262
      });                                                                                                    // 263
    }                                                                                                        // 264
    if (identity && identity.unverifiedEmail) {                                                              // 265
      unverifiedEmails[identity.unverifiedEmail] = true;                                                     // 266
    }                                                                                                        // 267
  });                                                                                                        // 268
                                                                                                             // 269
  var result = [];                                                                                           // 270
  _.keys(verifiedEmails).map(function (email) {                                                              // 271
    result.push({email: email,                                                                               // 272
                 verified: true,                                                                             // 273
                 primary: email === user.primaryEmail});                                                     // 274
  });                                                                                                        // 275
  _.keys(unverifiedEmails).map(function (email) {                                                            // 276
    if (!(email in verifiedEmails)) { result.push({email: email, verified: false}) }                         // 277
  });                                                                                                        // 278
                                                                                                             // 279
  // If `user.primaryEmail` is not among the verified emails, mark the first as primary.                     // 280
  if (!(user.primaryEmail in verifiedEmails) && result.length > 0) {                                         // 281
    result[0].primary = true;                                                                                // 282
  }                                                                                                          // 283
  return result;                                                                                             // 284
}                                                                                                            // 285
                                                                                                             // 286
SandstormDb.prototype.addContact = function addContact (ownerId, identityId) {                               // 287
  check(ownerId, String);                                                                                    // 288
  check(identityId, String);                                                                                 // 289
  var db = this;                                                                                             // 290
  var profile = db.getIdentity(identityId).profile;                                                          // 291
  db.collections.contacts.upsert({ownerId: ownerId, identityId: identityId}, {                               // 292
    ownerId: ownerId,                                                                                        // 293
    petname: profile && profile.name,                                                                        // 294
    created: new Date(),                                                                                     // 295
    identityId: identityId,                                                                                  // 296
  });                                                                                                        // 297
}                                                                                                            // 298
                                                                                                             // 299
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/user.js                                                                             //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var Crypto = Npm.require("crypto");                                                                          // 17
var Future = Npm.require("fibers/future");                                                                   // 18
                                                                                                             // 19
userPictureUrl = function (user) {                                                                           // 20
  if (user.services && !(user.profile && user.profile.picture)) {                                            // 21
    // Try to determine user's avatar URL from login service.                                                // 22
                                                                                                             // 23
    var google = user.services.google;                                                                       // 24
    if (google && google.picture) {                                                                          // 25
      return google.picture;                                                                                 // 26
    }                                                                                                        // 27
                                                                                                             // 28
    var github = user.services.github;                                                                       // 29
    if (github && github.id) {                                                                               // 30
      return "https://avatars.githubusercontent.com/u/" + github.id;                                         // 31
    }                                                                                                        // 32
                                                                                                             // 33
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be            // 34
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                // 35
    // Github are different because they are actually the identity providers, so they already know           // 36
    // the user logged in.                                                                                   // 37
  }                                                                                                          // 38
}                                                                                                            // 39
                                                                                                             // 40
fetchPicture = function (url) {                                                                              // 41
  try {                                                                                                      // 42
    var result = HTTP.get(url, {                                                                             // 43
      npmRequestOptions: { encoding: null },                                                                 // 44
      timeout: 5000                                                                                          // 45
    });                                                                                                      // 46
                                                                                                             // 47
    var metadata = {};                                                                                       // 48
                                                                                                             // 49
    metadata.mimeType = result.headers["content-type"];                                                      // 50
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                              // 51
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                             // 52
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                        // 53
    }                                                                                                        // 54
    var enc = result.headers["content-encoding"];                                                            // 55
    if (enc && enc !== "identity") {                                                                         // 56
      metadata.encoding = enc;                                                                               // 57
    }                                                                                                        // 58
                                                                                                             // 59
    return addStaticAsset(metadata, result.content);                                                         // 60
  } catch (err) {                                                                                            // 61
    console.error("failed to fetch user profile picture:", url, err.stack);                                  // 62
  }                                                                                                          // 63
}                                                                                                            // 64
                                                                                                             // 65
var ValidHandle = Match.Where(function (handle) {                                                            // 66
  check(handle, String);                                                                                     // 67
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                               // 68
});                                                                                                          // 69
                                                                                                             // 70
Accounts.onCreateUser(function (options, user) {                                                             // 71
  if (user.loginIdentities) {                                                                                // 72
    // it's an account                                                                                       // 73
    check(user, {_id: String,                                                                                // 74
                 createdAt: Date,                                                                            // 75
                 isAdmin: Match.Optional(Boolean),                                                           // 76
                 hasCompletedSignup: Match.Optional(Boolean),                                                // 77
                 signupKey: Match.Optional(String),                                                          // 78
                 signupNote: Match.Optional(String),                                                         // 79
                 signupEmail: Match.Optional(String),                                                        // 80
                 expires: Match.Optional(Date),                                                              // 81
                 appDemoId: Match.Optional(String),                                                          // 82
                 loginIdentities: [{id: String}],                                                            // 83
                 nonloginIdentities: [{id: String}]});                                                       // 84
                                                                                                             // 85
    if (Meteor.settings.public.quotaEnabled) {                                                               // 86
      user.experiments = user.experiments || {};                                                             // 87
      user.experiments = {                                                                                   // 88
        firstTimeBillingPrompt: Math.random() < 0.5 ? "control" : "test"                                     // 89
      };                                                                                                     // 90
    }                                                                                                        // 91
                                                                                                             // 92
    return user;                                                                                             // 93
  }                                                                                                          // 94
                                                                                                             // 95
  // Check profile.                                                                                          // 96
  if (options.profile) {                                                                                     // 97
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                     // 98
    check(options.profile, Match.ObjectIncluding({                                                           // 99
      name: Match.OneOf(null, Match.Optional(String)),                                                       // 100
      handle: Match.Optional(ValidHandle),                                                                   // 101
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                            // 102
    }));                                                                                                     // 103
  }                                                                                                          // 104
                                                                                                             // 105
  check(options.unverifiedEmail, Match.Optional(String));                                                    // 106
                                                                                                             // 107
  if (options.unverifiedEmail) {                                                                             // 108
    user.unverifiedEmail = options.unverifiedEmail;                                                          // 109
  }                                                                                                          // 110
  user.profile = _.pick(options.profile || {}, "name", "handle", "pronouns");                                // 111
                                                                                                             // 112
  // Try downloading avatar.                                                                                 // 113
  var url = userPictureUrl(user);                                                                            // 114
  if (url) {                                                                                                 // 115
    var assetId = fetchPicture(url);                                                                         // 116
    if (assetId) {                                                                                           // 117
      user.profile.picture = assetId;                                                                        // 118
    }                                                                                                        // 119
  }                                                                                                          // 120
                                                                                                             // 121
  var serviceUserId;                                                                                         // 122
  if (user.services && user.services.dev) {                                                                  // 123
    check(user.services.dev, {name: String, isAdmin: Boolean, hasCompletedSignup: Boolean});                 // 124
    serviceUserId = user.services.dev.name;                                                                  // 125
    user.profile.service = "dev";                                                                            // 126
  } else if ("expires" in user) {                                                                            // 127
    serviceUserId = user._id;                                                                                // 128
    user.profile.service = "demo";                                                                           // 129
  } else if (user.services && user.services.email) {                                                         // 130
    check(user.services.email,                                                                               // 131
          {email: String,                                                                                    // 132
           tokens: [{digest: String, algorithm: String, createdAt: Date}]});                                 // 133
    serviceUserId = user.services.email.email;                                                               // 134
    user.profile.service = "email";                                                                          // 135
  } else if (user.services && "google" in user.services) {                                                   // 136
    serviceUserId = user.services.google.id;                                                                 // 137
    user.profile.service = "google";                                                                         // 138
  } else if (user.services && "github" in user.services) {                                                   // 139
    serviceUserId = user.services.github.id;                                                                 // 140
    user.profile.service = "github";                                                                         // 141
  } else {                                                                                                   // 142
    throw new Meteor.Error(400, "user does not have a recognized identity provider: " +                      // 143
                           JSON.stringify(user));                                                            // 144
  }                                                                                                          // 145
  user._id = Crypto.createHash("sha256")                                                                     // 146
    .update(user.profile.service + ":" + serviceUserId).digest("hex");                                       // 147
                                                                                                             // 148
  return user;                                                                                               // 149
});                                                                                                          // 150
                                                                                                             // 151
// TODO delete this obsolete index.                                                                          // 152
Meteor.users._ensureIndex("identities.id", {unique: 1, sparse: 1});                                          // 153
                                                                                                             // 154
Meteor.users._ensureIndex("loginIdentities.id", {unique: 1, sparse: 1});                                     // 155
Meteor.users._ensureIndex("nonloginIdentities.id", {sparse: 1});                                             // 156
                                                                                                             // 157
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/migrations.js                                                                       //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                          // 1
//                                                                                                           // 2
// Because migrations can experience partial failure and likely have                                         // 3
// side-effects, we should be careful to make sure all migrations are                                        // 4
// idempotent and safe to accidentally run multiple times.                                                   // 5
                                                                                                             // 6
var Future = Npm.require("fibers/future");                                                                   // 7
                                                                                                             // 8
var updateLoginStyleToRedirect = function() {                                                                // 9
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 10
  ["google", "github"].forEach(function(serviceName) {                                                       // 11
    var config = configurations.findOne({service: serviceName});                                             // 12
    if (config && config.loginStyle !== "redirect") {                                                        // 13
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                       // 14
    }                                                                                                        // 15
  });                                                                                                        // 16
};                                                                                                           // 17
                                                                                                             // 18
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                                 // 19
  // In the before time, Google and Github login were enabled by default.                                    // 20
  //                                                                                                         // 21
  // This actually didn't make much sense, required the first user to configure                              // 22
  // OAuth, and had some trust-the-first-user properties that weren't totally                                // 23
  // secure.                                                                                                 // 24
  //                                                                                                         // 25
  // Now, we have admin-token, so we wish to disable all logins by default                                   // 26
  // (since they need to be configured anyway) but since users may have never                                // 27
  // explicitly told Sandstorm that Google or Github login should be enabled,                                // 28
  // we can't just use the value in the Settings collection, since it might                                  // 29
  // never have been set.                                                                                    // 30
  //                                                                                                         // 31
  // Thus, if the service is configured but absent from Settings, we should                                  // 32
  // explicitly enable it in Settings, and then the rest of the logic can just                               // 33
  // depend on what value is in Settings and default to false without breaking                               // 34
  // user installations.                                                                                     // 35
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 36
  ["google", "github"].forEach(function(serviceName) {                                                       // 37
    var config = configurations.findOne({service: serviceName});                                             // 38
    var serviceConfig = Settings.findOne({_id: serviceName});                                                // 39
    if (config && !serviceConfig) {                                                                          // 40
      // Only explicitly enable the login service if:                                                        // 41
      // 1) the service is already configured                                                                // 42
      // 2) there is no sandstorm configuration already present (the user was                                // 43
      //    using the previous default behavior).                                                            // 44
      Settings.insert({_id: serviceName, value: true});                                                      // 45
    }                                                                                                        // 46
  });                                                                                                        // 47
};                                                                                                           // 48
                                                                                                             // 49
var denormalizeInviteInfo = function() {                                                                     // 50
  // When a user is invited via a signup token, the `signupKey` field of their user table entry              // 51
  // has always been populated to indicate the key they used. This points into the SignupKeys table          // 52
  // which has more information about the key, namely a freeform note entered by the admin when              // 53
  // they created the key. In the case that the email invite form was used, the note has the form            // 54
  // "E-mail invite to <address>".                                                                           // 55
  //                                                                                                         // 56
  // Later, we decided it was useful to indicate in the users table visible to the admin                     // 57
  // information about the invite terms. Namely, for email invites we want to show the address               // 58
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                // 59
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing              // 60
  // from the SignupKeys table.                                                                              // 61
                                                                                                             // 62
  Meteor.users.find().forEach(function (user) {                                                              // 63
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {              // 64
      var signupInfo = SignupKeys.findOne(user.signupKey);                                                   // 65
      if (signupInfo && signupInfo.note) {                                                                   // 66
        var newFields = { signupNote: signupInfo.note };                                                     // 67
                                                                                                             // 68
        var prefix = "E-mail invite to ";                                                                    // 69
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                     // 70
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                      // 71
        }                                                                                                    // 72
                                                                                                             // 73
        Meteor.users.update(user._id, {$set: newFields});                                                    // 74
      }                                                                                                      // 75
    }                                                                                                        // 76
  });                                                                                                        // 77
}                                                                                                            // 78
                                                                                                             // 79
function mergeRoleAssignmentsIntoApiTokens() {                                                               // 80
  RoleAssignments.find().forEach(function(roleAssignment) {                                                  // 81
    ApiTokens.insert({                                                                                       // 82
      grainId: roleAssignment.grainId,                                                                       // 83
      userId: roleAssignment.sharer,                                                                         // 84
      roleAssignment: roleAssignment.roleAssignment,                                                         // 85
      petname: roleAssignment.petname,                                                                       // 86
      created: roleAssignment.created,                                                                       // 87
      owner: {user: {userId: roleAssignment.recipient,                                                       // 88
                     title: roleAssignment.title}},                                                          // 89
    });                                                                                                      // 90
  });                                                                                                        // 91
}                                                                                                            // 92
                                                                                                             // 93
function fixOasisStorageUsageStats() {}                                                                      // 94
// This migration only pertained to Oasis and it was successfully applied there. Since it referred           // 95
// to some global variables that we later wanted to remove and/or rename, we've since replaced it            // 96
// with a no-op.                                                                                             // 97
                                                                                                             // 98
function fetchProfilePictures() {                                                                            // 99
  Meteor.users.find({}).forEach(function (user) {                                                            // 100
    var url = userPictureUrl(user);                                                                          // 101
    if (url) {                                                                                               // 102
      console.log("Fetching user picture:", url);                                                            // 103
      var assetId = fetchPicture(url);                                                                       // 104
      if (assetId) {                                                                                         // 105
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                                 // 106
      }                                                                                                      // 107
    }                                                                                                        // 108
  });                                                                                                        // 109
}                                                                                                            // 110
                                                                                                             // 111
function assignPlans() {                                                                                     // 112
  if (Meteor.settings.public.quotaEnabled && SandstormDb.paymentsMigrationHook) {                            // 113
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                                     // 114
  }                                                                                                          // 115
}                                                                                                            // 116
                                                                                                             // 117
function removeKeyrings() {                                                                                  // 118
  // These blobs full of public keys were not intended to find their way into mongo and while                // 119
  // harmless they slow things down because they're huge. Remove them.                                       // 120
  Packages.update({"manifest.metadata.pgpKeyring": {$exists: true}},                                         // 121
      {$unset: {"manifest.metadata.pgpKeyring": ""}},                                                        // 122
      {multi: true});                                                                                        // 123
}                                                                                                            // 124
                                                                                                             // 125
function useLocalizedTextInUserActions() {                                                                   // 126
  function toLocalizedText(newObj, oldObj, field) {                                                          // 127
    if (field in oldObj) {                                                                                   // 128
      if (typeof oldObj[field] === "string") {                                                               // 129
        newObj[field] = {defaultText: oldObj[field]};                                                        // 130
      } else {                                                                                               // 131
        newObj[field] = oldObj[field];                                                                       // 132
      }                                                                                                      // 133
    }                                                                                                        // 134
  }                                                                                                          // 135
  UserActions.find({}).forEach(function (userAction) {                                                       // 136
    var fields = {};                                                                                         // 137
    toLocalizedText(fields, userAction, "appTitle");                                                         // 138
    toLocalizedText(fields, userAction, "title");                                                            // 139
    toLocalizedText(fields, userAction, "nounPhrase");                                                       // 140
    UserActions.update(userAction._id, {$set: fields});                                                      // 141
  });                                                                                                        // 142
}                                                                                                            // 143
                                                                                                             // 144
function verifyAllPgpSignatures() {                                                                          // 145
  Packages.find({}).forEach(function (pkg) {                                                                 // 146
    try {                                                                                                    // 147
      console.log("checking PGP signature for package:", pkg._id);                                           // 148
      var info = waitPromise(globalBackend.cap().tryGetPackage(pkg._id));                                    // 149
      if (info.authorPgpKeyFingerprint) {                                                                    // 150
        console.log("  " + info.authorPgpKeyFingerprint);                                                    // 151
        Packages.update(pkg._id,                                                                             // 152
            {$set: {authorPgpKeyFingerprint: info.authorPgpKeyFingerprint}});                                // 153
      } else {                                                                                               // 154
        console.log("  no signature");                                                                       // 155
      }                                                                                                      // 156
    } catch (err) {                                                                                          // 157
      console.error(err.stack);                                                                              // 158
    }                                                                                                        // 159
  });                                                                                                        // 160
}                                                                                                            // 161
                                                                                                             // 162
function splitUserIdsIntoAccountIdsAndIdentityIds() {                                                        // 163
  var Crypto = Npm.require("crypto");                                                                        // 164
  Meteor.users.find().forEach(function (user) {                                                              // 165
    var identity = {};                                                                                       // 166
    var serviceUserId;                                                                                       // 167
    if ("devName" in user) {                                                                                 // 168
      identity.service = "dev";                                                                              // 169
      serviceUserId = user.devName;                                                                          // 170
    } else if ("expires" in user) {                                                                          // 171
      identity.service = "demo";                                                                             // 172
      serviceUserId = user._id;                                                                              // 173
    } else if (user.services && "google" in user.services) {                                                 // 174
      identity.service = "google";                                                                           // 175
      if (user.services.google.email && user.services.google.verified_email) {                               // 176
        identity.verifiedEmail = user.services.google.email;                                                 // 177
      }                                                                                                      // 178
      serviceUserId = user.services.google.id;                                                               // 179
    } else if (user.services && "github" in user.services) {                                                 // 180
      identity.service = "github";                                                                           // 181
      identity.unverifiedEmail = user.services.github.email;                                                 // 182
      serviceUserId = user.services.github.id;                                                               // 183
    } else if (user.services && "emailToken" in user.services) {                                             // 184
      identity.service = "emailToken";                                                                       // 185
      identity.verifiedEmail = user.services.emailToken.email;                                               // 186
      serviceUserId = user.services.emailToken.email;                                                        // 187
    }                                                                                                        // 188
                                                                                                             // 189
    identity.id = Crypto.createHash("sha256")                                                                // 190
        .update(identity.service + ":" + serviceUserId).digest("hex");                                       // 191
                                                                                                             // 192
    if (user.profile) {                                                                                      // 193
      if (user.profile.name) {                                                                               // 194
        identity.name = user.profile.name;                                                                   // 195
      }                                                                                                      // 196
      if (user.profile.handle) {                                                                             // 197
        identity.handle = user.profile.handle;                                                               // 198
      }                                                                                                      // 199
      if (user.profile.picture) {                                                                            // 200
        identity.picture = user.profile.picture;                                                             // 201
      }                                                                                                      // 202
      if (user.profile.pronoun) {                                                                            // 203
        identity.pronoun = user.profile.pronoun;                                                             // 204
      }                                                                                                      // 205
      if (user.profile.email) {                                                                              // 206
        identity.unverifiedEmail = user.profile.email;                                                       // 207
      }                                                                                                      // 208
    }                                                                                                        // 209
    identity.main = true;                                                                                    // 210
                                                                                                             // 211
    Meteor.users.update(user._id, {$set: {identities: [identity]}});                                         // 212
                                                                                                             // 213
    Grains.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                     // 214
    Sessions.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                   // 215
    ApiTokens.update({userId: user._id},                                                                     // 216
                     {$set: {identityId: identity.id}},                                                      // 217
                     {multi: true});                                                                         // 218
    ApiTokens.update({"owner.user.userId": user._id},                                                        // 219
                     {$set: {"owner.user.identityId": identity.id}},                                         // 220
                     {multi: true});                                                                         // 221
    ApiTokens.update({"owner.grain.introducerUser": user._id},                                               // 222
                     {$set: {"owner.grain.introducerIdentity": identity.id}},                                // 223
                     {multi: true});                                                                         // 224
                                                                                                             // 225
    while (ApiTokens.update({"requirements.permissionsHeld.userId": user._id},                               // 226
                            {$set: {"requirements.$.permissionsHeld.identityId": identity.id},               // 227
                             $unset: {"requirements.$.permissionsHeld.userId": 1}},                          // 228
                            {multi: true}) > 0);                                                             // 229
    // The `$` operatorer modifies the first element in the array that matches the query. Since              // 230
    // there may be many matches, we need to repeat until no documents are modified.                         // 231
                                                                                                             // 232
  });                                                                                                        // 233
                                                                                                             // 234
  ApiTokens.remove({userInfo: {$exists: true}});                                                             // 235
  // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                    // 236
  // that this field could show up in the database was in this deprecated, no-longer-functional              // 237
  // form of API token.                                                                                      // 238
}                                                                                                            // 239
                                                                                                             // 240
function appUpdateSettings() {                                                                               // 241
  Settings.insert({_id: "appMarketUrl", value: "https://apps.sandstorm.io"});                                // 242
  Settings.insert({_id: "appIndexUrl", value: "https://app-index.sandstorm.io"});                            // 243
  Settings.insert({_id: "appUpdatesEnabled", value: true});                                                  // 244
}                                                                                                            // 245
                                                                                                             // 246
function moveDevAndEmailLoginDataIntoIdentities() {                                                          // 247
  Meteor.users.find().forEach(function (user) {                                                              // 248
    if (user.identities.length != 1) {                                                                       // 249
      throw new Error("User does not have exactly one identity: ", user);                                    // 250
    }                                                                                                        // 251
    var identity = user.identities[0];                                                                       // 252
    if (Match.test(identity.service, Object)) { return; } // Already migrated.                               // 253
                                                                                                             // 254
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");         // 255
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                          // 256
                                                                                                             // 257
    var serviceObject = {};                                                                                  // 258
    var fieldsToUnset = {}                                                                                   // 259
                                                                                                             // 260
    if (identity.service === "dev") {                                                                        // 261
      serviceObject.name = user.devName;                                                                     // 262
      fieldsToUnset.devName = 1;                                                                             // 263
    } else if (identity.service === "emailToken") {                                                          // 264
      serviceObject.tokens = user.services.emailToken.tokens;                                                // 265
      serviceObject.email = user.services.emailToken.email;                                                  // 266
      fieldsToUnset["services.emailToken"] = 1;                                                              // 267
    }                                                                                                        // 268
                                                                                                             // 269
    newIdentity.service = {};                                                                                // 270
    newIdentity.service[identity.service] = serviceObject;                                                   // 271
                                                                                                             // 272
    var modifier = {$set: {identities: [newIdentity]}};                                                      // 273
    if (Object.keys(fieldsToUnset).length > 0) {                                                             // 274
      modifier["$unset"] = fieldsToUnset;                                                                    // 275
    }                                                                                                        // 276
                                                                                                             // 277
    Meteor.users.update({_id: user._id}, modifier);                                                          // 278
  });                                                                                                        // 279
}                                                                                                            // 280
                                                                                                             // 281
function repairEmailIdentityIds() {                                                                          // 282
  var Crypto = Npm.require("crypto");                                                                        // 283
  Meteor.users.find({"identities.service.emailToken": {$exists: 1}}).forEach(function (user) {               // 284
    if (user.identities.length != 1) {                                                                       // 285
      throw new Error("User does not have exactly one identity: ", user);                                    // 286
    }                                                                                                        // 287
    var identity = user.identities[0];                                                                       // 288
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail",                 // 289
                             "profile");                                                                     // 290
    newIdentity.service = {email: identity.service.emailToken};                                              // 291
    newIdentity.id = Crypto.createHash("sha256")                                                             // 292
      .update("email:" + identity.service.emailToken.email).digest("hex");                                   // 293
                                                                                                             // 294
    Grains.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});           // 295
    Sessions.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});         // 296
    ApiTokens.update({identityId: identity.id},                                                              // 297
                     {$set: {identityId: newIdentity.id}},                                                   // 298
                     {multi: true});                                                                         // 299
    ApiTokens.update({"owner.user.identityId": identity.id},                                                 // 300
                     {$set: {"owner.user.identityId": newIdentity.id}},                                      // 301
                     {multi: true});                                                                         // 302
    ApiTokens.update({"owner.grain.introducerIdentity": identity.id},                                        // 303
                     {$set: {"owner.grain.introducerIdentity": newIdentity.id}},                             // 304
                     {multi: true});                                                                         // 305
                                                                                                             // 306
    while (ApiTokens.update({"requirements.permissionsHeld.identityId": identity.id},                        // 307
                            {$set: {"requirements.$.permissionsHeld.identityId": newIdentity.id}},           // 308
                            {multi: true}) > 0);                                                             // 309
                                                                                                             // 310
    Meteor.users.update({_id: user._id}, {$set: {identities: [newIdentity]}});                               // 311
  });                                                                                                        // 312
}                                                                                                            // 313
                                                                                                             // 314
function splitAccountUsersAndIdentityUsers() {                                                               // 315
  Meteor.users.find({"identities": {$exists: true}}).forEach(function (user) {                               // 316
    if (user.identities.length != 1) {                                                                       // 317
      throw new Error("User does not have exactly one identity: ", user);                                    // 318
    }                                                                                                        // 319
    var identity = user.identities[0];                                                                       // 320
    var identityUser = _.pick(user, "createdAt", "lastActive", "expires")                                    // 321
    identityUser._id = identity.id;                                                                          // 322
    identityUser.profile = identity.profile;                                                                 // 323
    _.extend(identityUser, _.pick(identity, "unverifiedEmail"));                                             // 324
    identityUser.profile.service = Object.keys(identity.service)[0];                                         // 325
                                                                                                             // 326
    // Updating this user needs to be a two step process because the `services` field typically              // 327
    // contains subfields that are constrained to be unique by Mongo indices.                                // 328
    identityUser.stagedServices = _.omit(user.services, "resume");                                           // 329
    if (identity.service.dev) {                                                                              // 330
      identityUser.stagedServices.dev = identity.service.dev;                                                // 331
    } else if (identity.service.email) {                                                                     // 332
      identityUser.stagedServices.email = identity.service.email;                                            // 333
    }                                                                                                        // 334
                                                                                                             // 335
    var accountUser = _.pick(user, "_id", "createdAt", "lastActive", "expires",                              // 336
                             "isAdmin", "signupKey", "signupNote", "signupEmail",                            // 337
                             "plan", "storageUsage", "isAppDemoUser", "appDemoId",                           // 338
                             "payments", "dailySentMailCount", "hasCompletedSignup");                        // 339
    accountUser.loginIdentities = [_.pick(identity, "id")];                                                  // 340
    accountUser.nonloginIdentities = [];                                                                     // 341
    if (user.services && user.services.resume) {                                                             // 342
      accountUser.services = {resume: user.services.resume};                                                 // 343
    }                                                                                                        // 344
                                                                                                             // 345
    accountUser.stashedOldUser = user;                                                                       // 346
                                                                                                             // 347
    ApiTokens.update({identityId: identityUser._id}, {$set: {accountId: user._id}},                          // 348
                     {multi: true});                                                                         // 349
                                                                                                             // 350
    Meteor.users.upsert({_id: identityUser._id}, identityUser);                                              // 351
    Meteor.users.update({_id: user._id}, accountUser);                                                       // 352
  });                                                                                                        // 353
                                                                                                             // 354
  Meteor.users.find({"stagedServices": {$exists: true}}).forEach(function (identity) {                       // 355
    Meteor.users.update({_id: identity._id}, {$unset: {stagedServices: 1},                                   // 356
                                              $set: {services: identity.stagedServices}});                   // 357
  });                                                                                                        // 358
}                                                                                                            // 359
                                                                                                             // 360
function populateContactsFromApiTokens() {                                                                   // 361
  ApiTokens.find({"owner.user.identityId": {$exists: 1},                                                     // 362
                  accountId: {$exists: 1}}).forEach(function(token) {                                        // 363
    var identityId = token.owner.user.identityId;                                                            // 364
    var profile = SandstormDb.prototype.getIdentity(identityId).profile;                                     // 365
    Contacts.upsert({ownerId: token.accountId, identityId: identityId}, {                                    // 366
      ownerId: token.accountId,                                                                              // 367
      petname: profile && profile.name,                                                                      // 368
      created: new Date(),                                                                                   // 369
      identityId: identityId,                                                                                // 370
      profile: profile,                                                                                      // 371
    });                                                                                                      // 372
  });                                                                                                        // 373
}                                                                                                            // 374
                                                                                                             // 375
// This must come after all the functions named within are defined.                                          // 376
// Only append to this list!  Do not modify or remove list entries;                                          // 377
// doing so is likely change the meaning and semantics of user databases.                                    // 378
var MIGRATIONS = [                                                                                           // 379
  updateLoginStyleToRedirect,                                                                                // 380
  enableLegacyOAuthProvidersIfNotInSettings,                                                                 // 381
  denormalizeInviteInfo,                                                                                     // 382
  mergeRoleAssignmentsIntoApiTokens,                                                                         // 383
  fixOasisStorageUsageStats,                                                                                 // 384
  fetchProfilePictures,                                                                                      // 385
  assignPlans,                                                                                               // 386
  removeKeyrings,                                                                                            // 387
  useLocalizedTextInUserActions,                                                                             // 388
  verifyAllPgpSignatures,                                                                                    // 389
  splitUserIdsIntoAccountIdsAndIdentityIds,                                                                  // 390
  appUpdateSettings,                                                                                         // 391
  moveDevAndEmailLoginDataIntoIdentities,                                                                    // 392
  repairEmailIdentityIds,                                                                                    // 393
  splitAccountUsersAndIdentityUsers,                                                                         // 394
  populateContactsFromApiTokens,                                                                             // 395
];                                                                                                           // 396
                                                                                                             // 397
function migrateToLatest() {                                                                                 // 398
  if (Meteor.settings.replicaNumber) {                                                                       // 399
    // This is a replica. Wait for the first replica to perform migrations.                                  // 400
                                                                                                             // 401
    console.log("Waiting for migrations on replica zero...");                                                // 402
                                                                                                             // 403
    var done = new Future();                                                                                 // 404
    var change = function (doc) {                                                                            // 405
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                   // 406
      if (doc.value >= MIGRATIONS.length) done.return();                                                     // 407
    }                                                                                                        // 408
    var observer = Migrations.find({_id: "migrations_applied"}).observe({                                    // 409
      added: change,                                                                                         // 410
      changed: change                                                                                        // 411
    });                                                                                                      // 412
                                                                                                             // 413
    done.wait();                                                                                             // 414
    observer.stop();                                                                                         // 415
    console.log("Migrations have completed on replica zero.");                                               // 416
                                                                                                             // 417
  } else {                                                                                                   // 418
    var applied = Migrations.findOne({_id: "migrations_applied"});                                           // 419
    var start;                                                                                               // 420
    if (!applied) {                                                                                          // 421
      // Migrations table is not yet seeded with a value.  This means it has                                 // 422
      // applied 0 migrations.  Persist this.                                                                // 423
      Migrations.insert({_id: "migrations_applied", value: 0});                                              // 424
      start = 0;                                                                                             // 425
    } else {                                                                                                 // 426
      start = applied.value;                                                                                 // 427
    }                                                                                                        // 428
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                           // 429
                                                                                                             // 430
    for (var i = start ; i < MIGRATIONS.length ; i++) {                                                      // 431
      // Apply migration i, then record that migration i was successfully run.                               // 432
      console.log("Applying migration " + (i+1));                                                            // 433
      MIGRATIONS[i]();                                                                                       // 434
      Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                                  // 435
      console.log("Applied migration " + (i+1));                                                             // 436
    }                                                                                                        // 437
  }                                                                                                          // 438
}                                                                                                            // 439
                                                                                                             // 440
SandstormDb.prototype.migrateToLatest = migrateToLatest;                                                     // 441
                                                                                                             // 442
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
