(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var check = Package.check.check;
var Match = Package.check.Match;
var Random = Package.random.Random;
var _ = Package.underscore._;
var SandstormDb = Package['sandstorm-db'].SandstormDb;

/* Package-scope variables */
var SandstormPermissions;

(function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                         //
// packages/sandstorm-permissions/permissions.js                                                           //
//                                                                                                         //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                           //
// Sandstorm - Personal Cloud Sandbox                                                                      // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                   // 2
// All rights reserved.                                                                                    // 3
//                                                                                                         // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                         // 5
// you may not use this file except in compliance with the License.                                        // 6
// You may obtain a copy of the License at                                                                 // 7
//                                                                                                         // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                            // 9
//                                                                                                         // 10
// Unless required by applicable law or agreed to in writing, software                                     // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                       // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                // 13
// See the License for the specific language governing permissions and                                     // 14
// limitations under the License.                                                                          // 15
                                                                                                           // 16
var Crypto = Npm.require("crypto");                                                                        // 17
var Url = Npm.require("url");                                                                              // 18
                                                                                                           // 19
SandstormPermissions = {};                                                                                 // 20
                                                                                                           // 21
function PermissionSet(array) {                                                                            // 22
  // A wrapper around an array of booleans.                                                                // 23
                                                                                                           // 24
  if (!array) {                                                                                            // 25
    this.array = [];                                                                                       // 26
  } else if (array instanceof Array) {                                                                     // 27
    this.array = array.slice(0);                                                                           // 28
  } else {                                                                                                 // 29
    throw new Error("don't know how to interpret as PermissionSet: " + array);                             // 30
  }                                                                                                        // 31
}                                                                                                          // 32
                                                                                                           // 33
// Methods for mutating a PermissionSet by combining it with another PermissionSet.                        // 34
// These return a boolean indicating whether the operation had any effect.                                 // 35
                                                                                                           // 36
PermissionSet.prototype.add = function(other) {                                                            // 37
  var changed = false;                                                                                     // 38
  for (var ii = 0; ii < other.array.length; ++ii) {                                                        // 39
    var old = this.array[ii];                                                                              // 40
    this.array[ii] = this.array[ii] || other.array[ii];                                                    // 41
    if (old != this.array[ii]) {                                                                           // 42
      changed = true;                                                                                      // 43
    }                                                                                                      // 44
  }                                                                                                        // 45
  return changed;                                                                                          // 46
}                                                                                                          // 47
                                                                                                           // 48
PermissionSet.prototype.remove = function(other) {                                                         // 49
  var changed = false;                                                                                     // 50
  for (var ii = 0; ii < other.array.length && ii < this.array.length; ++ii) {                              // 51
    var old = this.array[ii];                                                                              // 52
    this.array[ii] = this.array[ii] && !other.array[ii];                                                   // 53
    if (old != this.array[ii]) {                                                                           // 54
      changed = true;                                                                                      // 55
    }                                                                                                      // 56
  }                                                                                                        // 57
  return changed;                                                                                          // 58
}                                                                                                          // 59
                                                                                                           // 60
PermissionSet.prototype.intersect = function(other) {                                                      // 61
  var changed = false;                                                                                     // 62
  for (var ii = 0; ii < this.array.length; ++ii) {                                                         // 63
    var old = this.array[ii];                                                                              // 64
    this.array[ii] = this.array[ii] && other.array[ii];                                                    // 65
    if (old != this.array[ii]) {                                                                           // 66
      changed = true;                                                                                      // 67
    }                                                                                                      // 68
  }                                                                                                        // 69
  return changed;                                                                                          // 70
}                                                                                                          // 71
                                                                                                           // 72
function roleAssignmentPermissions(roleAssignment, viewInfo) {                                             // 73
  var result = new PermissionSet([]);                                                                      // 74
                                                                                                           // 75
  if (!roleAssignment || "none" in roleAssignment) {                                                       // 76
    if (viewInfo.roles) {                                                                                  // 77
      for (var ii = 0; ii < viewInfo.roles.length; ++ii) {                                                 // 78
        var roleDef = viewInfo.roles[ii];                                                                  // 79
        if (roleDef.default) {                                                                             // 80
          result = new PermissionSet(roleDef.permissions);                                                 // 81
          break;                                                                                           // 82
        }                                                                                                  // 83
      }                                                                                                    // 84
    }                                                                                                      // 85
  } else if ("allAccess" in roleAssignment) {                                                              // 86
    var length = 0;                                                                                        // 87
    if (viewInfo.permissions) {                                                                            // 88
      length = viewInfo.permissions.length;                                                                // 89
    }                                                                                                      // 90
    var array = new Array(length);                                                                         // 91
    for (var ii = 0; ii < array.length; ++ii) {                                                            // 92
      array[ii] = true;                                                                                    // 93
    }                                                                                                      // 94
    result = new PermissionSet(array);                                                                     // 95
  } else if ("roleId" in roleAssignment && viewInfo.roles && viewInfo.roles.length > 0) {                  // 96
    var roleDef = viewInfo.roles[roleAssignment.roleId];                                                   // 97
    if (roleDef) {                                                                                         // 98
      result = new PermissionSet(roleDef.permissions);                                                     // 99
    }                                                                                                      // 100
  }                                                                                                        // 101
                                                                                                           // 102
  if (roleAssignment) {                                                                                    // 103
    result.add(new PermissionSet(roleAssignment.addPermissionSet));                                        // 104
    result.remove(new PermissionSet(roleAssignment.removePermissionSet));                                  // 105
  }                                                                                                        // 106
  return result;                                                                                           // 107
}                                                                                                          // 108
                                                                                                           // 109
function collectEdges(db, vertex) {                                                                        // 110
  // Given a vertex in the sharing graph in the format specified by the `check()` invocation below,        // 111
  // collects the data needed for permissions computations pertaining to that vertex. There are three      // 112
  // self-explanatory special cases for the return value. In order of decreasing precedence, they          // 113
  // are: `{grainDoesNotExist: true}`, `{grainIsPublic: true}`, and `{disallowedAnonymousAccess: true}`.   // 114
  // In all other cases, this function returns an object of the form:                                      // 115
  //   `{edgesByRecipient: <object>, terminalEdge: Optional(<object>)}`.                                   // 116
  // The `edgesByRecipient` field is a map that coalesces chains of `parentToken` UiView tokens into       // 117
  // direct user-to-user edges; its keys are identity IDs of recipient users and its values are lists      // 118
  // of "edge" objects of the form                                                                         // 119
  //   `{sharer: OneOf(<identityId>, "OwningAccount"), roleAssignments: <list of role assignments>}`.      // 120
  // The role assignments should be applied in sequence to compute the set of permissions that flow to a   // 121
  // recipient from a sharer. The `terminalEdge` field is an edge object representing the link to          // 122
  // `vertex` from the nearest user in the sharing graph. If `vertex` is already a user, then this edge is
  // trivial and its `roleAssignments` field is an empty list. If `terminalEdge` is not present,           // 124
  // then there is no such link.                                                                           // 125
  //                                                                                                       // 126
  // TODO(someday): Once UiView tokens can have membrane requirements, we'll need to account for           // 127
  // them in this computation.                                                                             // 128
  check(vertex,                                                                                            // 129
        Match.OneOf({token: Match.ObjectIncluding({_id: String, grainId: String})},                        // 130
                    {grain: Match.ObjectIncluding(                                                         // 131
                      {_id: String,                                                                        // 132
                       identityId: Match.OneOf(String, null, undefined)})}));                              // 133
                                                                                                           // 134
  var grainId;                                                                                             // 135
  if (vertex.token) {                                                                                      // 136
    grainId = vertex.token.grainId;                                                                        // 137
  } else if (vertex.grain) {                                                                               // 138
    grainId = vertex.grain._id;                                                                            // 139
  }                                                                                                        // 140
  var grain = db.getGrain(grainId);                                                                        // 141
  if (!grain) {                                                                                            // 142
    return {grainDoesNotExist: true};                                                                      // 143
  }                                                                                                        // 144
                                                                                                           // 145
  if (!grain.private) {                                                                                    // 146
    return {grainIsPublic: true};                                                                          // 147
  } else if (vertex.grain && !vertex.grain.identityId) {                                                   // 148
    return {disallowedAnonymousAccess: true};                                                              // 149
  }                                                                                                        // 150
                                                                                                           // 151
  var result = {edgesByRecipient: {}};                                                                     // 152
                                                                                                           // 153
  var tokensById = {};                                                                                     // 154
  db.collections.apiTokens.find({grainId: grainId,                                                         // 155
                                 revoked: {$ne: true}}).forEach(function(token) {                          // 156
    tokensById[token._id] = token;                                                                         // 157
  });                                                                                                      // 158
                                                                                                           // 159
  function computeEdge(token) {                                                                            // 160
    var roleAssignments = [];                                                                              // 161
    var curToken = token;                                                                                  // 162
    while (curToken && curToken.parentToken) {                                                             // 163
      // For a `parentToken` provider, no role assignment means don't attenuate.                           // 164
      if (curToken.roleAssignment) {                                                                       // 165
        roleAssignments.push(curToken.roleAssignment);                                                     // 166
      }                                                                                                    // 167
      curToken = tokensById[curToken.parentToken];                                                         // 168
    }                                                                                                      // 169
    if (curToken && curToken.grainId && curToken.identityId) {                                             // 170
      roleAssignments.push(curToken.roleAssignment);                                                       // 171
      return {sharer: curToken.identityId, roleAssignments: roleAssignments};                              // 172
    } else {                                                                                               // 173
      return;                                                                                              // 174
    }                                                                                                      // 175
  }                                                                                                        // 176
                                                                                                           // 177
  if (vertex.token) {                                                                                      // 178
    result.terminalEdge = computeEdge(tokensById[vertex.token._id]);                                       // 179
  } else if (vertex.grain) {                                                                               // 180
    result.terminalEdge = {sharer: vertex.grain.identityId, roleAssignments: []};                          // 181
  }                                                                                                        // 182
                                                                                                           // 183
  for (var id in tokensById) {                                                                             // 184
    var token = tokensById[id];                                                                            // 185
    if (Match.test(token.owner, {user: Match.Any})) {                                                      // 186
      var edge = computeEdge(token);                                                                       // 187
      if (edge) {                                                                                          // 188
        var recipient = token.owner.user.identityId ;                                                      // 189
        if (!result.edgesByRecipient[recipient]) {                                                         // 190
          result.edgesByRecipient[recipient] = [];                                                         // 191
        }                                                                                                  // 192
        result.edgesByRecipient[recipient].push(edge);                                                     // 193
      }                                                                                                    // 194
    }                                                                                                      // 195
  }                                                                                                        // 196
                                                                                                           // 197
  var owningUser = Meteor.users.findOne({_id: grain.userId});                                              // 198
  if (owningUser) {                                                                                        // 199
    SandstormDb.getUserIdentityIds(owningUser).forEach(function(identityId) {                              // 200
      result.edgesByRecipient[identityId] = [{sharer: "OwningAccount", roleAssignments: []}];              // 201
    });                                                                                                    // 202
  }                                                                                                        // 203
                                                                                                           // 204
  return result;                                                                                           // 205
}                                                                                                          // 206
                                                                                                           // 207
SandstormPermissions.grainPermissions = function(db, vertex, viewInfo) {                                   // 208
  // Computes the permissions of a vertex. If the vertex is not allowed to open the grain,                 // 209
  // returns null. Otherwise, returns an array of bools representing the permissions held.                 // 210
  var edges = collectEdges(db, vertex);                                                                    // 211
  if (edges.openerIsOwner) {                                                                               // 212
    return roleAssignmentPermissions({allAccess: null}, viewInfo).array;                                   // 213
  }                                                                                                        // 214
  if (edges.grainIsPublic) {                                                                               // 215
    // Grains using the old sharing model always share the default role to anyone who has the              // 216
    // grain URL.                                                                                          // 217
    return roleAssignmentPermissions({none: null}, viewInfo).array;                                        // 218
  }                                                                                                        // 219
  if (edges.grainDoesNotExist || !edges.terminalEdge || edges.disallowedAnonymousAccess) {                 // 220
    return null;                                                                                           // 221
  }                                                                                                        // 222
                                                                                                           // 223
  var openerIdentityId = edges.terminalEdge.sharer;                                                        // 224
  var edgesByRecipient = edges.edgesByRecipient;                                                           // 225
  var owner = "OwningAccount";                                                                             // 226
                                                                                                           // 227
  var permissionsMap = {};                                                                                 // 228
  // Keeps track of the permissions that the opener receives from each user. The final result of           // 229
  // our computation will be stored in permissionsMap[owner].                                              // 230
                                                                                                           // 231
  var userStack = [openerIdentityId];                                                                      // 232
                                                                                                           // 233
  var openerAttenuation = roleAssignmentPermissions({allAccess: null}, viewInfo);                          // 234
  edges.terminalEdge.roleAssignments.forEach(function(roleAssignment) {                                    // 235
    openerAttenuation.intersect(roleAssignmentPermissions(roleAssignment, viewInfo));                      // 236
  });                                                                                                      // 237
  permissionsMap[openerIdentityId] = openerAttenuation;                                                    // 238
                                                                                                           // 239
  while (userStack.length > 0) {                                                                           // 240
    var recipient = userStack.pop();                                                                       // 241
    if (edgesByRecipient[recipient]) {                                                                     // 242
      edgesByRecipient[recipient].forEach(function (inEdge) {                                              // 243
        var sharer = inEdge.sharer;                                                                        // 244
        var needToPush = false;                                                                            // 245
        if (!permissionsMap[sharer]) {                                                                     // 246
          permissionsMap[sharer] = new PermissionSet();                                                    // 247
          needToPush = true;                                                                               // 248
        }                                                                                                  // 249
                                                                                                           // 250
        var newPermissions = roleAssignmentPermissions({allAccess: null}, viewInfo);                       // 251
        inEdge.roleAssignments.forEach(function(roleAssignment) {                                          // 252
          newPermissions.intersect(roleAssignmentPermissions(roleAssignment, viewInfo));                   // 253
        });                                                                                                // 254
        newPermissions.intersect(permissionsMap[recipient]);                                               // 255
                                                                                                           // 256
        // Optimization: we don't care about permissions that we've already proven the opener has.         // 257
        if (permissionsMap[owner]) {                                                                       // 258
          newPermissions.remove(permissionsMap[owner]);                                                    // 259
        }                                                                                                  // 260
                                                                                                           // 261
        if (permissionsMap[sharer].add(newPermissions)) {                                                  // 262
          needToPush = true;                                                                               // 263
        }                                                                                                  // 264
        if (needToPush) {                                                                                  // 265
          userStack.push(sharer);                                                                          // 266
        }                                                                                                  // 267
      });                                                                                                  // 268
    }                                                                                                      // 269
  }                                                                                                        // 270
  if (permissionsMap[owner]) {                                                                             // 271
    return permissionsMap[owner].array;                                                                    // 272
  } else {                                                                                                 // 273
    return null;                                                                                           // 274
  }                                                                                                        // 275
}                                                                                                          // 276
                                                                                                           // 277
SandstormPermissions.mayOpenGrain = function(db, vertex) {                                                 // 278
  // Determines whether the vertex is allowed to open the grain by searching depth first                   // 279
  // for a path of active role assignments leading from the grain owner to the vertex.                     // 280
                                                                                                           // 281
  var edges = collectEdges(db, vertex);                                                                    // 282
  if (edges.grainDoesNotExist || edges.disallowedAnonymousAccess) {                                        // 283
    return false;                                                                                          // 284
  }                                                                                                        // 285
  if (edges.grainIsPublic) {                                                                               // 286
    return true;                                                                                           // 287
  }                                                                                                        // 288
  var edgesByRecipient = edges.edgesByRecipient;                                                           // 289
  var owner = "OwningAccount";                                                                             // 290
  if (!edges.terminalEdge) { return false; }                                                               // 291
  if (owner == edges.terminalEdge.sharer) {                                                                // 292
    return true;                                                                                           // 293
  }                                                                                                        // 294
                                                                                                           // 295
  var openerIdentityId = edges.terminalEdge.sharer;                                                        // 296
  var stackedUsers = {};                                                                                   // 297
  stackedUsers[openerIdentityId] = true;                                                                   // 298
  var userStack = [openerIdentityId];                                                                      // 299
                                                                                                           // 300
  while (userStack.length > 0) {                                                                           // 301
    var recipient = userStack.pop();                                                                       // 302
    var edges = edgesByRecipient[recipient];                                                               // 303
    if (edges) {                                                                                           // 304
      for (var ii = 0; ii < edges.length; ++ii) {                                                          // 305
        var inEdge = edges[ii];                                                                            // 306
        var sharer = inEdge.sharer;                                                                        // 307
        if (sharer == owner) {                                                                             // 308
          return true;                                                                                     // 309
        }                                                                                                  // 310
        if (!stackedUsers[sharer]) {                                                                       // 311
          userStack.push(sharer);                                                                          // 312
          stackedUsers[sharer] = true;                                                                     // 313
        }                                                                                                  // 314
      }                                                                                                    // 315
    }                                                                                                      // 316
  }                                                                                                        // 317
  return false;                                                                                            // 318
}                                                                                                          // 319
                                                                                                           // 320
SandstormPermissions.downstreamTokens = function(db, root) {                                               // 321
  // Computes a list of the UiView tokens that are downstream in the sharing graph from a given            // 322
  // source. The source, `root`, can either be a token or a (grain, user) pair. The exact format           // 323
  // of `root` is specified in the `check()` invocation below.                                             // 324
  //                                                                                                       // 325
  // TODO(someday): Once UiView tokens can have membrane requirements, we'll need to account for           // 326
  // them in this computation.                                                                             // 327
                                                                                                           // 328
  check(root, Match.OneOf({token: Match.ObjectIncluding({_id: String, grainId: String})},                  // 329
                          {grain: Match.ObjectIncluding({_id: String, identityId: String})}));             // 330
                                                                                                           // 331
  var result = [];                                                                                         // 332
  var tokenStack = [];                                                                                     // 333
  var stackedTokens = {};                                                                                  // 334
  var tokensBySharer = {};                                                                                 // 335
  var tokensByParent = {};                                                                                 // 336
  var tokensById = {};                                                                                     // 337
                                                                                                           // 338
  function addChildren(tokenId) {                                                                          // 339
    var children = tokensByParent[tokenId];                                                                // 340
    if (children) {                                                                                        // 341
      children.forEach(function (child) {                                                                  // 342
        if (!stackedTokens[child._id]) {                                                                   // 343
          tokenStack.push(child);                                                                          // 344
          stackedTokens[child._id] = true;                                                                 // 345
        }                                                                                                  // 346
      });                                                                                                  // 347
    }                                                                                                      // 348
  }                                                                                                        // 349
                                                                                                           // 350
  function addSharedTokens(sharer) {                                                                       // 351
    var sharedTokens = tokensBySharer[sharer];                                                             // 352
    if (sharedTokens) {                                                                                    // 353
      sharedTokens.forEach(function (sharedToken) {                                                        // 354
        if (!stackedTokens[sharedToken._id]) {                                                             // 355
          tokenStack.push(sharedToken);                                                                    // 356
          stackedTokens[sharedToken._id] = true;                                                           // 357
        }                                                                                                  // 358
      });                                                                                                  // 359
    }                                                                                                      // 360
  }                                                                                                        // 361
                                                                                                           // 362
  var grainId;                                                                                             // 363
  if (root.token) {                                                                                        // 364
    grainId = root.token.grainId;                                                                          // 365
  } else if (root.grain) {                                                                                 // 366
    grainId = root.grain._id;                                                                              // 367
  }                                                                                                        // 368
                                                                                                           // 369
  var grain = db.getGrain(grainId);                                                                        // 370
  if (!grain || !grain.private ) { return result; }                                                        // 371
                                                                                                           // 372
  db.collections.apiTokens.find({grainId: grainId,                                                         // 373
                                 revoked: {$ne: true}}).forEach(function (token) {                         // 374
    tokensById[token._id] = token;                                                                         // 375
    if (token.parentToken) {                                                                               // 376
      if (!tokensByParent[token.parentToken]) {                                                            // 377
        tokensByParent[token.parentToken] = [];                                                            // 378
      }                                                                                                    // 379
      tokensByParent[token.parentToken].push(token);                                                       // 380
    } else if (token.identityId) {                                                                         // 381
      if (!tokensBySharer[token.identityId]) {                                                             // 382
        tokensBySharer[token.identityId] = [];                                                             // 383
      }                                                                                                    // 384
      tokensBySharer[token.identityId].push(token);                                                        // 385
    }                                                                                                      // 386
  });                                                                                                      // 387
                                                                                                           // 388
  if (root.token) {                                                                                        // 389
    addChildren(root.token._id);                                                                           // 390
  } else if (root.grain) {                                                                                 // 391
    addSharedTokens(root.grain.identityId);                                                                // 392
  }                                                                                                        // 393
                                                                                                           // 394
  while (tokenStack.length > 0) {                                                                          // 395
    var token = tokenStack.pop();                                                                          // 396
    result.push(token);                                                                                    // 397
    addChildren(token._id);                                                                                // 398
    if (token.owner && token.owner.user) {                                                                 // 399
      addSharedTokens(token.owner.user.identityId);                                                        // 400
    }                                                                                                      // 401
  }                                                                                                        // 402
                                                                                                           // 403
  return result;                                                                                           // 404
}                                                                                                          // 405
                                                                                                           // 406
SandstormPermissions.createNewApiToken = function (db, provider, grainId, petname,                         // 407
                                                   roleAssignment, owner) {                                // 408
  // Creates a new UiView API token. If `rawParentToken` is set, creates a child token.                    // 409
  check(grainId, String);                                                                                  // 410
  check(petname, String);                                                                                  // 411
  check(roleAssignment, db.roleAssignmentPattern);                                                         // 412
  // Meteor bug #3877: we get null here instead of undefined when we                                       // 413
  // explicitly pass in undefined.                                                                         // 414
  check(provider, Match.OneOf({identityId: String, accountId: String},                                     // 415
                              {rawParentToken: String}));                                                  // 416
  check(owner, Match.OneOf({webkey: {forSharing: Boolean,                                                  // 417
                                     expiresIfUnusedDuration: Match.Optional(Number)}},                    // 418
                           {user: {identityId: String,                                                     // 419
                                   title: String}}));                                                      // 420
                                                                                                           // 421
  var grain = db.getGrain(grainId);                                                                        // 422
  if (!grain) {                                                                                            // 423
    throw new Meteor.Error(403, "Unauthorized", "No grain found.");                                        // 424
  }                                                                                                        // 425
                                                                                                           // 426
  var token = Random.secret();                                                                             // 427
  var apiToken = {                                                                                         // 428
    _id: Crypto.createHash("sha256").update(token).digest("base64"),                                       // 429
    grainId: grainId,                                                                                      // 430
    roleAssignment: roleAssignment,                                                                        // 431
    petname: petname,                                                                                      // 432
    created: new Date(),                                                                                   // 433
    expires: null,                                                                                         // 434
  };                                                                                                       // 435
                                                                                                           // 436
  var parentForSharing = false;                                                                            // 437
  if (provider.rawParentToken) {                                                                           // 438
    var parentToken = Crypto.createHash("sha256").update(provider.rawParentToken).digest("base64");        // 439
    var parentApiToken = db.collections.apiTokens.findOne(                                                 // 440
      {_id: parentToken, grainId: grainId, objectId: {$exists: false}});                                   // 441
    if (!parentApiToken) {                                                                                 // 442
      throw new Meteor.Error(403, "No such parent token found.");                                          // 443
    }                                                                                                      // 444
    if (parentApiToken.forSharing) {                                                                       // 445
      parentForSharing = true;                                                                             // 446
    }                                                                                                      // 447
                                                                                                           // 448
    // TODO(soon): For a while this field was failing to get set due to a typo. We should migrate          // 449
    //   old entries in ApiTokens to populate this field where appropriate.                                // 450
    apiToken.identityId = parentApiToken.identityId;                                                       // 451
    apiToken.accountId = parentApiToken.accountId;                                                         // 452
                                                                                                           // 453
    apiToken.parentToken = parentToken;                                                                    // 454
  } else if (provider.identityId) {                                                                        // 455
    apiToken.identityId = provider.identityId;                                                             // 456
    apiToken.accountId = provider.accountId;                                                               // 457
  }                                                                                                        // 458
                                                                                                           // 459
  if (owner.webkey) {                                                                                      // 460
    apiToken.owner = {webkey: null};                                                                       // 461
    apiToken.forSharing = parentForSharing || owner.webkey.forSharing;                                     // 462
    if (owner.webkey.expiresIfUnusedDuration) {                                                            // 463
      apiToken.expiresIfUnused = new Date(Date.now() + owner.webkey.expiresIfUnusedDuration);              // 464
    }                                                                                                      // 465
  } else if (owner.user) {                                                                                 // 466
    var pkg = db.collections.packages.findOne(grain.packageId);                                            // 467
    var appTitle = (pkg && pkg.manifest && pkg.manifest.appTitle) || { defaultText: ""};                   // 468
    var grainInfo = {appTitle: appTitle};                                                                  // 469
                                                                                                           // 470
    if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {                     // 471
      var icons = pkg.manifest.metadata.icons;                                                             // 472
      grainInfo.icon = icons.grain || icons.appGrid;                                                       // 473
    }                                                                                                      // 474
    // Only provide an app ID if we have no icon asset to provide and need to offer an identicon.          // 475
    if (!grainInfo.icon && pkg) {                                                                          // 476
      grainInfo.appId = pkg.appId;                                                                         // 477
    }                                                                                                      // 478
    apiToken.owner = {                                                                                     // 479
      user: {                                                                                              // 480
        identityId: owner.user.identityId,                                                                 // 481
        title: owner.user.title,                                                                           // 482
        // lastUsed: ??                                                                                    // 483
        denormalizedGrainMetadata: grainInfo,                                                              // 484
      }                                                                                                    // 485
    };                                                                                                     // 486
  }                                                                                                        // 487
                                                                                                           // 488
  db.collections.apiTokens.insert(apiToken);                                                               // 489
                                                                                                           // 490
  var endpointUrl = Url.parse(process.env.ROOT_URL).protocol + "//" + db.makeWildcardHost("api");          // 491
  return {id: apiToken._id, token: token, endpointUrl: endpointUrl, parentApiToken: parentApiToken};       // 492
}                                                                                                          // 493
                                                                                                           // 494
// Make self-destructing tokens actually self-destruct, so they don't                                      // 495
// clutter the token list view.                                                                            // 496
SandstormPermissions.cleanupSelfDestructing = function (db) {                                              // 497
  return function () {                                                                                     // 498
    var now = new Date();                                                                                  // 499
    db.collections.apiTokens.remove({expiresIfUnused: {$lt: now}});                                        // 500
  }                                                                                                        // 501
}                                                                                                          // 502
                                                                                                           // 503
Meteor.methods({                                                                                           // 504
  transitiveShares: function(identityId, grainId) {                                                        // 505
    check(identityId, String);                                                                             // 506
    check(grainId, String);                                                                                // 507
    if (this.userId) {                                                                                     // 508
      var db = this.connection.sandstormDb;                                                                // 509
      return SandstormPermissions.downstreamTokens(db,                                                     // 510
          {grain: {_id: grainId, identityId: identityId }});                                               // 511
    }                                                                                                      // 512
  },                                                                                                       // 513
                                                                                                           // 514
  newApiToken: function (provider, grainId, petname, roleAssignment, owner) {                              // 515
    check(provider, Match.OneOf({identityId: String}, {rawParentToken: String}));                          // 516
    var db = this.connection.sandstormDb;                                                                  // 517
    if (provider.identityId) {                                                                             // 518
      if (!this.userId || !db.userHasIdentity(this.userId, provider.identityId)) {                         // 519
        throw new Meteor.Error(403, "Not an identity of the current user: " + provider.identityId);        // 520
      }                                                                                                    // 521
    }                                                                                                      // 522
    if (provider.identityId) {                                                                             // 523
      provider.accountId = this.userId;                                                                    // 524
    }                                                                                                      // 525
    return SandstormPermissions.createNewApiToken(                                                         // 526
      this.connection.sandstormDb, provider, grainId, petname, roleAssignment, owner);                     // 527
  },                                                                                                       // 528
                                                                                                           // 529
  updateApiToken: function (token, newFields) {                                                            // 530
    var db = this.connection.sandstormDb;                                                                  // 531
                                                                                                           // 532
    check(token, String);                                                                                  // 533
    check(newFields, {petname: Match.Optional(String),                                                     // 534
                      roleAssignment: Match.Optional(db.roleAssignmentPattern),                            // 535
                      revoked: Match.Optional(Boolean)});                                                  // 536
                                                                                                           // 537
    if (!this.userId) {                                                                                    // 538
      throw new Meteor.Error(403, "Must be logged in to modify a token");                                  // 539
    }                                                                                                      // 540
    var apiToken = db.collections.apiTokens.findOne(token);                                                // 541
    if (!apiToken) {                                                                                       // 542
      throw new Meteor.Error(404, "No such token found.");                                                 // 543
    }                                                                                                      // 544
                                                                                                           // 545
    if (db.userHasIdentity(this.userId, apiToken.identityId)) {                                            // 546
      var modifier = {$set: newFields};                                                                    // 547
      db.collections.apiTokens.update(token, modifier);                                                    // 548
    } else {                                                                                               // 549
      throw new Meteor.Error(403, "User not authorized to modify this token.");                            // 550
    }                                                                                                      // 551
  }                                                                                                        // 552
});                                                                                                        // 553
                                                                                                           // 554
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-permissions'] = {
  SandstormPermissions: SandstormPermissions
};

})();

//# sourceMappingURL=sandstorm-permissions.js.map
